from typing import Tuple


def parse_version_string(v: str) -> Tuple:
    """
    "2.4" --> (1, 2, 0)
    "2" --> (2, 0, 0)
    """
    parts = v.split(".")
    parts = (parts + 3 * ["0"])[:3]
    return tuple(int(x) for x in parts)



from pprint import pprint
from unittest import mock

from pycrunch.session.diagnostics import diagnostic_engine

def test_modules():
    actual = diagnostic_engine.get_modules()
    pprint(actual)

def test_diagnostic_env():
    with mock.patch('os.environ', {'test': 'var'}):
        env = dict(diagnostic_engine.get_env())
        assert env['test'] == 'var'

def test_summary():
    actual = diagnostic_engine.summary()
    pprint(actual)
    assert 'modules' in actual
    assert 'env' in actual


import logging
from asyncio import get_event_loop
from pathlib import Path

from pycrunch.api.serializers import serialize_test_set_state
from pycrunch.api.shared import file_watcher, pipe
from pycrunch.discovery.strategy import create_test_discovery
from pycrunch.introspection.history import execution_history
from pycrunch.session import config
from pycrunch.session.diagnostics import diagnostic_engine
from pycrunch.shared.models import all_tests

logger = logging.getLogger(__name__)

class EngineState:
    def __init__(self):
        # self.all_tests = dict()
        folder_auto = str(Path('.').absolute())
        logger.info(f'folder is: {folder_auto}')
        self.folder = folder_auto
        self.all_tests = all_tests
        self.runtime_configuration_ready = False

    async def will_start_test_discovery(self):
        #  TODO: How about running this automatically before engine is connected?
        self.prepare_runtime_configuration_if_necessary()
        self.begin_watch_for_config_changes()
        discovery_engine = create_test_discovery()
        test_set = discovery_engine.find_tests_in_folder(self.folder)
        await engine.test_discovery_will_become_available(test_set)

    def begin_watch_for_config_changes(self):
        config.watch_for_config_changes()

    async def will_start_diagnostics_collection(self):
        self.prepare_runtime_configuration_if_necessary()

        logger.info('will_start_diagnostics_collection')
        await pipe.push(event_type='diagnostics_did_become_available', engine=config.runtime_engine, **diagnostic_engine.summary())
        logger.info('diagnostics_did_become_available')

    async def will_send_timings(self):
        await pipe.push(event_type='execution_history_did_become_available', **execution_history.to_json(), folder=self.folder)

    async def test_discovery_will_become_available(self, test_set):
        """
        :type test_set: pycrunch.discovery.simple.TestSet
        """
        for discovered_test in test_set.tests:
            is_pinned = config.is_test_pinned(discovered_test.fqn)
            self.all_tests.test_discovered(discovered_test.fqn, discovered_test, is_pinned)

        # TODO: maybe do not wait for the signal from plugin to start discovery.
        self.all_tests.discard_tests_not_in_map()
        self.notify_clients_about_tests_change()
        logger.debug('discovery_did_become_available')
        logger.debug(f'Adding files for watch: total of {len(test_set.files)}')
        file_watcher.watch(test_set.files)

    def notify_clients_about_tests_change(self):
        logger.debug('notify_clients_about_tests_change')
        serialized_data = serialize_test_set_state(self.all_tests.tests)
        get_event_loop().create_task(
            pipe.push(event_type='discovery_did_become_available', **serialized_data, folder=self.folder)
        )

    async def tests_will_run(self, tests):
        logger.info('tests_will_run')

        for test in tests:
            self.all_tests.test_will_run(test.discovered_test.fqn)

        self.notify_clients_about_tests_change()

    async def tests_did_run(self, results):
        for k, v in results.items():
            self.all_tests.test_did_run(k, v)

        self.notify_clients_about_tests_change()

    async def tests_will_pin(self, fqns):
        for fqn in fqns:
            self.all_tests.pin_test(fqn)
        self.notify_clients_about_tests_change()

        self.save_pinned_state()


    async def tests_will_unpin(self, fqns):
        for fqn in fqns:
            self.all_tests.unpin_test(fqn)

        self.save_pinned_state()
        self.notify_clients_about_tests_change()

    def save_pinned_state(self):
        config.save_pinned_tests_config(self.all_tests.get_pinned_tests())

    def engine_mode_will_change(self, new_mode):
        config.runtime_mode_will_change(new_mode)

    def get_engine_mode(self):
        return config.engine_mode

    def prepare_runtime_configuration_if_necessary(self):
        if not self.runtime_configuration_ready:
            self.runtime_configuration_ready = True
            config.load_runtime_configuration()


engine = EngineState()






import logging
import os
import sys
import traceback

import pytest

from pycrunch.plugins.pytest_support.interception_plugin import PyTestInterceptionPlugin
from pycrunch.runner import _abstract_runner
from pycrunch.runner.execution_result import ExecutionResult

logger = logging.getLogger(__name__)


class PyTestRunnerEngine(_abstract_runner.Runner):
    def __init__(self, child_config):
        """

        :type child_config: pycrunch.child_runtime.child_config.ChildRuntimeConfig
        """
        self.child_config = child_config


    def run_test(self, test):
        """

        :type test: object
        :return val pycrunch.runner.execution_result.ExecutionResult
        """
        execution_result = ExecutionResult()
        try:
            fqn_test_to_run = test.filename + '::' + test.name

            plugin = PyTestInterceptionPlugin([fqn_test_to_run])

            # pytest.main(['tests_two.py::test_x', '-p', 'no:terminal'])
            # q - quite
            # s - do not capture console logs
            # l - show variable values in the current stack
            additional_pytest_args = ['-qs', '-l']
            plugins_arg = []

            if not self.child_config.load_pytest_plugins:
                os.environ['PYTEST_DISABLE_PLUGIN_AUTOLOAD'] = 'True'
                plugins_arg += ['-p', 'no:junitxml']

            # --trace-config
            # , '-p', 'no:helpconfig', - cannot be disabled
            all_args = additional_pytest_args + plugins_arg
            # print(all_args, file=sys.__stdout__)
            if self.child_config.enable_remote_debug:
                try:
                    # Todo: this is too late to check for debugger existence.
                    #   Need verify before `debug` button click
                    import pydevd_pycharm
                    pydevd_pycharm.settrace('127.0.0.1', suspend=False, port=self.child_config.remote_debug_port, stdoutToServer=True, stderrToServer=True)
                except ModuleNotFoundError as e:
                    print('---\nFailed to import pydevd_pycharm', file=sys.__stdout__)
                    print('  Make sure you install pudb pycharm bindings by running:', file=sys.__stdout__)
                    print('pip install pydevd-pycharm\n---', file=sys.__stdout__)
                    raise
            pytest.main([fqn_test_to_run] + all_args, plugins=[plugin])

            # pytest.main([fqn_test_to_run, '-qs'], plugins=[plugin])

            if plugin.guess_run_status(test.name):
                execution_result.run_did_succeed()
            else:
                execution_result.run_did_fail()
            # x.stop()
            # logger.debug('after exec_module')
        except Exception as e:
            etype, value, current_traceback = sys.exc_info()
            execution_result.record_exception(etype=etype, value=value, current_traceback=current_traceback)
            execution_result.run_did_fail()
            last_call = -1
            traceback.print_exception(etype=etype, value=value, tb=current_traceback, limit=last_call, file=sys.stdout)
            # traceback.print_exc(file=sys.stdout)
            # print(str(e))
            logger.exception('Error while executing _run_test', exc_info=e)
        return execution_result




from unittest import TestCase

class TestX():
    def test_1(self):
        pass


class SomeClassInhereted(TestCase):
    def test_1(self):
        pass


class AbstractMyTestCase(TestCase):
    pass


class SomeClassDoublyInherited(AbstractMyTestCase):
    def test_1(self):
        pass



import json
import unittest
import uuid
import pprint
from unittest.mock import MagicMock

from pycrunch.api.serializers import CoverageRun
from pycrunch.child_runtime.coverage_hal import CoverageAbstraction
from pycrunch.insights.variables_inspection import InsightTimeline
from pycrunch.introspection.clock import Clock
from pycrunch.runner.execution_result import ExecutionResult
from pycrunch.shared.primitives import TestMetadata


class TestCoverageRun(unittest.TestCase):
    def test_sample(self):
        test_meta = self.sample_metadata()
        state_timeline = InsightTimeline(clock=Clock())

        execution_result = ExecutionResult()
        execution_result.run_did_succeed()
        execution_result.state_timeline_did_become_available(state_timeline)
        run = CoverageRun('test1', 1, test_meta, execution_result)
        pprint.pprint(run.as_json())

    def sample_metadata(self):
        test_meta = TestMetadata('file1', 'test_a', 'module_a', 'module_a:test_a', 'queued')
        return test_meta

    def test_unsupported_type_uuid_in_state_timeline_dict(self):
        state_timeline = InsightTimeline(clock=Clock())
        state_timeline.start()

        state_timeline.record(dict(some=uuid.uuid1()))
        t = dict(a=1, nested=dict(some=uuid.uuid1()))
        x = pprint.pformat(t)

        state_timeline.record(t)
        state_timeline.record(TestCoverageRun)
        state_timeline.record(self)

        execution_result = ExecutionResult()
        execution_result.run_did_succeed()
        execution_result.state_timeline_did_become_available(state_timeline)
        x = json.dumps(state_timeline.as_json())
        print(x)

    def test_coverage_lib_version_4(self):
        sut = self.create_sut(is_v5_or_greater=False)

        assert 'data_file' not in sut.get_coverage_arguments()

    def test_coverage_lib_version_5(self):
        sut = self.create_sut(is_v5_or_greater=True)

        actual = sut.get_coverage_arguments()

        assert 'data_file' in actual
        assert actual['data_file'] == None

    def create_sut(self, is_v5_or_greater):
        sut = CoverageAbstraction(True, None, None)
        sut.is_coverage_v5_or_greater = MagicMock(return_value=is_v5_or_greater)
        return sut



import asyncio
import logging
import pickle
from queue import Empty, Queue

from pycrunch.introspection.history import execution_history
from pycrunch.networking.protocol_state import ProtocolState
from pycrunch.scheduling.messages import ScheduledTaskDefinitionMessage
from pycrunch.session import config

logger = logging.getLogger(__name__)


class TestRunnerServerProtocol(asyncio.Protocol):
    # 1 object - 1 connection
    def __init__(self, tasks, completion_future, timeline):
        self.timeline = timeline
        self.completion_future = completion_future
        self.tasks = tasks
        self.transport = None
        # Will be determined after handshake
        self.task_id = None
        self.message_queue = Queue()
        self.results = []
        self.need_more_data = False
        self.message_length = 0
        self.read_so_far = 0
        self.message_buffer = None
        self.state_machine = ProtocolState(self.message_queue)

    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        logger.debug(f'Connection from {peername}')
        self.transport = transport

    def feed_datagram(self, data):
        self.state_machine.feed(data)

    def data_received(self, data):
        self.feed_datagram(data)
        self.process_messages()

    def process_messages(self):
        while True:
            msg = self.try_get_next_message()
            if not msg:
                break

            self.process_single_message(msg)


    def process_single_message(self, msg):
        logger.info(f'process_single_message - {msg.kind}')
        if msg.kind == 'handshake':
            found_task = self.find_task_with_id(msg)
            if found_task is None:
                raise Exception('no task found for subprocess. ')

            logger.debug(f'sending task definition, {found_task.id}')
            msg_to_reply = ScheduledTaskDefinitionMessage(task=found_task, coverage_exclusions=config.coverage_exclusions)
            bytes_msg = pickle.dumps(msg_to_reply)
            self.transport.write(bytes_msg)
        if msg.kind == 'test_run_results':
            results = msg.results
            self.timeline.mark_event('TCP: Got test run results from subprocess')
            self.results_did_become_available(results)
        if msg.kind == 'timings':
            self.timeline.mark_event('TCP: Got timings from subprocess')
            execution_history.save(msg.timeline)
        if msg.kind == 'close':
            self.timeline.mark_event('TCP: Received close message')
            logger.debug('Close the client socket')
            self.transport.close()
            self.completion_future.set_result(self.results)

    def try_get_next_message(self):
        msg = None
        try:
            msg = self.message_queue.get_nowait()
        except Empty as e:
            pass
            # print('process_messages: nothing to process')
        return msg

    def connection_lost(self, ex = None):
        if not self.completion_future.done():
            self.completion_future.set_result(self.results)

    def find_task_with_id(self, msg):
        found_task = None
        for x in self.tasks:
            if x.id == msg.task_id:
                found_task = x
                break
        return found_task

    def results_did_become_available(self, results):
        self.results = results

    def force_close(self):
        self.transport.close()
        self.completion_future.cancel()




from pycrunch.discovery.ast_discovery import AstTestDiscovery
# from pycrunch.discovery.simple import SimpleTestDiscovery


def create_test_discovery(root_directory=None, configuration=None):
    return AstTestDiscovery(root_directory, configuration)
    # or
    # return SimpleTestDiscovery(root_directory, configuration)



class AbstractMessage():
    kind = 'abstract'

class HandshakeMessage(AbstractMessage):
    def __init__(self, task_id):
        self.task_id = task_id
        self.kind = 'handshake'

class ScheduledTaskDefinitionMessage(AbstractMessage):
    def __init__(self, task, coverage_exclusions):
        self.task = task
        self.coverage_exclusions = coverage_exclusions  # type: list[str]
        self.kind = 'test-run-task'

class TestResultsAvailableMessage(AbstractMessage):
    def __init__(self, results):
        self.results = results
        self.kind = 'test_run_results'


class TestRunTimingsMessage(AbstractMessage):
    def __init__(self, timeline):
        self.timeline = timeline
        self.kind = 'timings'


class CloseConnectionMessage(AbstractMessage):
    def __init__(self, task_id):
        self.task_id = task_id
        self.kind = 'close'




from unittest import mock
from unittest.mock import mock_open, call

from pycrunch.runner.execution_result import ExecutionResult
from pycrunch.session.configuration import Configuration
from pycrunch.session.state import engine
from pycrunch.shared.models import TestState, AllTests
from pycrunch.shared.primitives import TestMetadata


def test_pinned_test_should_change_status():
    sut = AllTests()
    fqn = 'module:test'
    state = TestState(TestMetadata('dummy', 'dummy', 'dummy', fqn, 'pending'), ExecutionResult(), False)
    sut.test_discovered(fqn, state, False)
    single_test = sut.tests.get(fqn, None)
    assert single_test.pinned == False
    sut.pin_test(fqn)
    single_test = sut.tests.get(fqn, None)
    assert single_test.pinned == True



read_data = '''
discovery:
  exclusions:
   - directory_1
   - directory_2
engine:
  runtime: simple
  mode: manual
'''
def test_pin_test_in_engine_should_store_configuration():
    sut = Configuration()
    with mock.patch('io.open', mock_open(read_data=read_data)) as mock_file:
        mock_write = mock_open()
        mock_write.return_value.write = None
        mock_file.side_effect =[
            mock_open(read_data=read_data).return_value,
            mock_open(mock_write).return_value
        ]

        sut.save_pinned_tests_config(['a', 'b'])
        mock_write.assert_has_calls([call().write("pinned-tests"),])



read_data_with_pinned = '''
discovery:
  exclusions:
   - directory_1
   - directory_2
engine:
  runtime: simple
  mode: manual
pinned-tests:
  - a:fqn
  - b
'''
def test_should_load_pinned_tests_from_configuration_on_startup():
    sut = Configuration()
    with mock.patch('io.open', mock_open(read_data=read_data_with_pinned)) as mock_file:
        sut.load_runtime_configuration()
        assert 'a:fqn' in sut.pinned_tests
        assert 'b' in sut.pinned_tests
        assert 'c' not in sut.pinned_tests



import asyncio
import logging

from pycrunch.api.shared import pipe
from pycrunch.execution_watchdog.tasks import AbstractWatchdogTask, TestExecutionBeginTask, TestExecutionEndTask
from pycrunch.execution_watchdog.watchdog_pipeline import watchdog_pipeline

logger = logging.getLogger(__name__)

# Blocks until the internal flag is true.
termination_event = asyncio.Event()


"""
  Responsible for debugger/execution 
    - stop run button visibility 
    - signal to stop received from plugin
    - Timeout deadline while waiting for tests
"""
class WatchdogDispatcher:
    def __init__(self):
        pass

    async def run_once(self):
        task = await watchdog_pipeline.get_task()
        await self.execute_task(task)

    async def execute_task(self, task: AbstractWatchdogTask):
        if task.name == 'watchdog_begin':
            task: TestExecutionBeginTask = task
            termination_event.clear()
            await pipe.push(
                event_type='watchdog_begin',
                total_tests=task.total_tests
            )
        if task.name == 'watchdog_end':
            task: TestExecutionEndTask = task
            termination_event.clear()
            await pipe.push(
                event_type='watchdog_end',
            )
        if task.name == 'watchdog_terminate':
            termination_event.set()


async def watchdog_dispather_thread():
    logger.info('WatchDog Dispatcher thread -- Start')
    dp = WatchdogDispatcher()

    while True:
        logger.debug('WatchDog Dispatcher thread -- inside event loop, waiting for task...')
        try:
            await dp.run_once()
        except asyncio.CancelledError:
            logger.warning('CancelledError while processing WatchDog Dispatcher queue')
            break
        except Exception as e:
            logger.exception('Exception in WatchDog Dispatcher Thread, ', exc_info=e)
            break

    logger.debug('WatchDog Dispatcher thread -- End')



import asyncio


async def run(engine_to_use, timeline, port, task_id):
    from pycrunch.child_runtime.client_protocol import EchoClientProtocol

    timeline.mark_event('Run: inside method, imports')
    import sys
    from pathlib import Path

    # from pprint import pprint
    # from multiprocessing.connection import Client
    # timeline.mark_event('Run: imported multiprocessing.connection')
    timeline.mark_event('Run: imported TestRunner')

    timeline.mark_event('Run: imports done')

    test_configuration = None
    tests_to_run = []
    # add root of django project
    sys.path.insert(0, str(Path('.').absolute()))
    # todo: make configurable instead
    # sys.path.insert(0, '/code')
    global loop

    on_con_lost = loop.create_future()
    timeline.mark_event('TCP: Opening connection')
    transport, protocol1 = await loop.create_connection(
        lambda: EchoClientProtocol(on_con_lost, task_id, timeline, engine_to_use),
        '127.0.0.1',
        port,
    )
    await on_con_lost


async def main():
    # import pydevd_pycharm
    #
    # pydevd_pycharm.settrace('localhost', port=21345, stdoutToServer=True, stderrToServer=True)
    from pycrunch.introspection.timings import Timeline

    timeline = Timeline('multiprocess run engine')
    timeline.start()
    timeline.mark_event('__main__')

    timeline.mark_event('ArgumentParser: Init')
    import argparse

    parser = argparse.ArgumentParser(description='PyCrunch CLI')

    parser.add_argument('--engine', help='Engine used, one of [pytest, django, simple]')
    parser.add_argument('--port', help='PyCrunch-Engine server port to connect')
    parser.add_argument(
        '--task-id', help='Id of task when multiple test runners ran at same time'
    )
    parser.add_argument(
        '--load-pytest-plugins', help='If this is true, execution will be slower.'
    )
    parser.add_argument(
        '--enable-remote-debug',
        action='store_true',
        help='If this is true, remote debug will be enabled on a --remote-debugger-port',
    )
    parser.add_argument(
        '--remote-debugger-port',
        help='If remote debug is enabled, this will specify a port used to connect to PyCharm pudb',
    )
    parser.add_argument(
        '--collect-perf',
        action='store_true',
        help='If this is enabled, timings/timelines will be collected and sent to web_ui for further look',
    )
    args = parser.parse_args()
    timeline.mark_event('ArgumentParser: parse_args completed')
    from pycrunch.child_runtime.child_config import child_config

    engine_to_use = args.engine
    if args.engine:
        child_config.use_engine(engine_to_use)
    if args.load_pytest_plugins.lower() == 'true':
        child_config.load_pytest_plugins = True
    if args.enable_remote_debug:
        child_config.enable_remote_debugging(args.remote_debugger_port)
    if args.collect_perf:
        child_config.enable_timing_collection()

    timeline.mark_event('Before run')

    await run(
        engine_to_use=engine_to_use,
        timeline=timeline,
        port=args.port,
        task_id=args.task_id,
    )


loop = asyncio.get_event_loop()

try:
    loop.run_until_complete(main())
    # loop.run_forever()
except KeyboardInterrupt:
    print('graceful quit!')
    pass



from unittest.mock import MagicMock

from pycrunch.insights.variables_inspection import InsightTimeline, inject_timeline
from pycrunch.insights import trace
from pycrunch.introspection.clock import clock


def test_should_trace_something_to_timeline():
    future_is_here = 1
    trace(future_is_here=future_is_here)
    trace(traced_variable=42)

def marker(name_of_the_marker):
    pass

def test_single_trace():
    trace(value_a=1)


def test_marker_single_variable_trace():
    trace(value_a=1)
    marker('point_of_interest')
    trace(value_a=42)
    trace(value_b='after sleep')

def test_marker_multiple_variables_trace():
    for i in range(10):
        trace(loop_counter=i)

    trace(value_a=__name__)
    marker('point_of_interest')
    trace(value_a=42)
    some_local = dict(
        x=1,
        y=2,
        size=dict(
            w=10,
            h=10),
        string='abc',
    )
    trace(some_local=some_local)


def test_inject_timeline_and_call_it_on_trace():
    future_is_here = 1
    mock = MagicMock()
    inject_timeline(mock)
    trace(future_is_here=future_is_here)
    mock.record.assert_called_with(future_is_here=future_is_here)

def test_timeline_one_variable():
    variable_name = 42
    x = InsightTimeline(clock)
    x.start()

    x.record(variable_name=42)

    first = x.variables[0]
    assert 'variable_name' == first.name
    assert 42 == first.value


def test_timeline_two_variables():
    variable_a = 42
    variable_b = 'Bret Victor'
    x = InsightTimeline(clock)
    x.start()
    x.record(variable_a=variable_a, brightest_mind=variable_b)

    first = x.variables[0]
    second = x.variables[1]
    assert 'variable_a' == first.name
    assert 'brightest_mind' == second.name
    assert 42 == first.value
    assert variable_b == second.value

def test_timeline_positional_arguments_only():
    variable_a = 42
    x = InsightTimeline(clock)
    x.start()
    x.record(variable_a, 777)

    first = x.variables[0]
    assert '1' == first.name
    assert 42 == first.value

    second = x.variables[1]
    assert '2' == second.name
    assert 777 == second.value



class ChildRuntimeConfig:
    def __init__(self):
        self.load_pytest_plugins = False
        self.enable_remote_debug = False
        self.remote_debug_port = -1
        self.runtime_engine = 'pytest'
        self.collect_timings = False

    def use_engine(self, new_engine):
        self.runtime_engine = new_engine

    def enable_remote_debugging(self, port_to_use):
        self.enable_remote_debug = True
        self.remote_debug_port = int(port_to_use)

    def enable_timing_collection(self):
        self.collect_timings = True


child_config = ChildRuntimeConfig()



import asyncio
import logging
from pathlib import Path
from typing import Set

from pycrunch.pipeline import execution_pipeline
from pycrunch.pipeline.file_removed_task import FileRemovedTask
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from ..constants import CONFIG_FILE_NAME
from ._abstract_watcher import Watcher

logger = logging.getLogger(__name__)


def create_handler(files_to_watch: "Set[str]", event_loop):
    from ..pipeline.file_modification_task import FileModifiedNotificationTask

    class CustomFSWatchHandler(FileSystemEventHandler):
        """Logs all the events captured."""

        def __init__(self):
            super().__init__()
            self.files_to_watch = files_to_watch
            self.event_loop = event_loop

        def should_watch_file(self, entry: 'str') -> bool:
            return entry.endswith(('.py', '.pyx', '.pyd', CONFIG_FILE_NAME))

        def known_file(self, file):
            return file in self.files_to_watch

        def on_moved(self, event):
            super().on_moved(event)

            if self.known_file(event.src_path):
                self.add_task_in_queue(FileRemovedTask(file=event.src_path))

            if not self.should_watch_file(event.src_path):
                return


            what = 'directory' if event.is_directory else 'file'
            logger.debug(
                "Moved %s: from %s to %s", what, event.src_path, event.dest_path
            )
            self.send_modification_message(event.dest_path, 'moved')

        def on_created(self, event):
            super().on_created(event)
            if not self.should_watch_file(event.src_path):
                return

            self.send_modification_message(event.src_path, 'created')

        def on_deleted(self, event):
            super().on_deleted(event)
            if not self.known_file(event.src_path):
                return

            if not self.should_watch_file(event.src_path):
                return

            self.add_task_in_queue(FileRemovedTask(file=event.src_path))
            logger.info('Added file removal for pipeline ' + event.src_path)

        def add_task_in_queue(self, t: "AbstractTask"):
            # Hack included: it is not possible to submit into asyncio queue from another thread, therefore:
            # https://stackoverflow.com/questions/59083275/simplest-way-to-put-an-item-in-an-asyncio-queue-from-sync-code-running-in-a-sepa
            # main event loop doesn't know about action made in thread
            fut = asyncio.run_coroutine_threadsafe(
                execution_pipeline.put_raw(t), self.event_loop
            )
            fut.result()

        def on_modified(self, event):
            super().on_modified(event)

            if not self.known_file(event.src_path):
                return

            if not self.should_watch_file(event.src_path):
                return

            self.send_modification_message(event.src_path, 'modified')

            what = 'directory' if event.is_directory else 'file'
            logger.debug("Modified %s: %s", what, event.src_path)

        def send_modification_message(self, filename, context):
            logger.debug('Adding file modification for processing ' + filename)

            self.add_task_in_queue(FileModifiedNotificationTask(file=filename, context=context))
            logger.debug(' -- done Added ' + filename)

    return CustomFSWatchHandler()


class FSWatcher(Watcher):
    def __init__(self):
        self._started = False
        self.files = set()
        self.event_loop = asyncio.get_event_loop()

    def watch(self, files):
        logger.debug('watch...')
        self.files.update(files)
        logger.debug(f"Total files to watch: {len(self.files)}")

        self.start_thread_if_not_running()

    def start_thread_if_not_running(self):
        if self._started:
            return

        logger.debug('start_thread_if_not_running->Creating fs_observer')
        path = str(Path('.').absolute())
        logger.debug('watch root path: ', path)
        observer = Observer()
        observer.schedule(
            create_handler(self.files, event_loop=self.event_loop),
            path=path,
            recursive=True,
        )
        observer.start()
        logger.info('Started watch thread...')
        self._started = True



import io

from pycrunch.api import shared
from pycrunch.pipeline.abstract_task import AbstractTask
from pycrunch.session import config

import logging

logger = logging.getLogger(__name__)

class DownloadFileTask(AbstractTask):
    def __init__(self, filename):
        self.filename = filename

    async def run(self):
        target_file = config.path_mapping.map_local_to_remote(self.filename)
        my_file = io.FileIO(target_file, 'r')
        content = my_file.read()
        logger.debug(f'sending file {self.filename}')
        await shared.pipe.push(event_type='file_did_load',
                         filename=self.filename,
                         file_contents=content,
                         )
        # return Response(content, mimetype='application/x-python-code')


class AbstractWatchdogTask:
    name: str
    def run(self):
        raise Exception('Abstract method call in AbstractWatchdogTask->run.')


class TestExecutionBeginTask(AbstractWatchdogTask):
    def __init__(self, total_tests: int):
        self.total_tests = total_tests
        self.name = 'watchdog_begin'


class TestExecutionEndTask(AbstractWatchdogTask):
    def __init__(self):
        self.name = 'watchdog_end'


class TerminateTestExecutionTask(AbstractWatchdogTask):
    def __init__(self):
        self.name = 'watchdog_terminate'



from pycrunch.session.configuration import PathMapping


def test_path_mapping_ok():
    sut = PathMapping('/code', '/gleb/ba/django')
    actual = sut.map_to_local_fs('/code/main.py')
    assert '/gleb/ba/django/main.py' == actual


def test_path_mapping_nested():
    sut = PathMapping('/code', '/gleb/ba/django')
    actual = sut.map_to_local_fs('/code/core/code/main.py')
    assert '/gleb/ba/django/core/code/main.py' == actual


from asyncio import Queue

import logging

from pycrunch.execution_watchdog.tasks import AbstractWatchdogTask

logger = logging.getLogger(__name__)


class WatchdogPipeline:
    def __init__(self):
        self.q = Queue()

    def add_task(self, task: AbstractWatchdogTask):
        logger.debug('Received task in queue')
        self.q.put_nowait(task)

    async def get_task(self):
        return await self.q.get()


watchdog_pipeline = WatchdogPipeline()



import asyncio
import pickle
import struct
import os

from pycrunch.child_runtime.test_runner import TestRunner
from pycrunch.scheduling.messages import CloseConnectionMessage, HandshakeMessage, TestResultsAvailableMessage, TestRunTimingsMessage

counter = 0


class EchoClientProtocol(asyncio.Protocol):
    """
    This class represents multiprocess-child connection client protocol
    Supported input actions are
       - test-run-task
    Output messages:
       - test_run_results
       - timings
       - close [asks server to close connection so child process can terminate]
    """
    def __init__(self, on_connection_lost, task_id, timeline, engine_to_use):
        self.engine_to_use = engine_to_use
        self.timeline = timeline
        self.task_id = task_id
        self.on_con_lost = on_connection_lost
        self.transport = None
        global counter
        self.connection_counter = counter+1
        counter += 1

    def connection_made(self, transport):
        self.timeline.mark_event(f'TCP: Connection established...')
        self.transport = transport
        msg = HandshakeMessage(self.task_id)
        msg_bytes = pickle.dumps(msg)
        # print(f'[{self.connection_counter}]Handshake sent: {msg.task_id}')
        self.send_with_header(msg_bytes)



    def data_received(self, data):
        msg = pickle.loads(data)
        if msg.kind == 'test-run-task':
            print(f'[{os.getpid()}] [task_id: {msg.task.id}] Data received: test-run-task;')
            timeline = self.timeline
            timeline.mark_event(f'TCP: Received tests to run; id: {msg.task.id}')

            runner_engine = None

            timeline.mark_event('Deciding on runner engine...')
            from pycrunch.plugins.pytest_support.pytest_runner_engine import PyTestRunnerEngine

            if self.engine_to_use == 'django':
                from pycrunch.session import config
                config.runtime_engine_will_change(self.engine_to_use)
                config.prepare_django()

            from pycrunch.child_runtime.child_config import child_config
            runner_engine = PyTestRunnerEngine(child_config)

            # should have env from pycrunch config here
            r = TestRunner(runner_engine, timeline, msg.coverage_exclusions, child_config)
            timeline.mark_event(f'Run: about to run tests')
            try:
                timeline.mark_event(f'Run: total tests planned: {len(msg.task.tests)}')
                results = r.run(msg.task.tests)
                timeline.mark_event('Run: Completed, sending results')

                msg = TestResultsAvailableMessage(results)
                bytes_to_send = self.safe_pickle(msg)
                self.send_with_header(bytes_to_send)
                timeline.mark_event('TCP: results sent')
            except Exception as e:
                import sys
                import traceback
                print('----! Unexpected exception during PyCrunch test execution:', file=sys.__stdout__)
                traceback.print_exc(file=sys.__stdout__)

                timeline.mark_event('Run: exception during execution')

            timeline.stop()

            if child_config.collect_timings:
                msg = TestRunTimingsMessage(timeline)
                bytes_to_send1 = pickle.dumps(msg)
                self.send_with_header(bytes_to_send1)

            msg = CloseConnectionMessage(self.task_id)
            bytes_to_send2 = pickle.dumps(msg)
            self.send_with_header(bytes_to_send2)

    def safe_pickle(self, msg):
        try:
           return pickle.dumps(msg)
        except Exception as e:
            for k, v in msg.results.items():
                v.execution_result.state_timeline.make_safe_for_pickle()

        return pickle.dumps(msg)

    def send_with_header(self, bytes_to_send):
        # we use format: {length}{payload} to deal with TCP coalescing
        length_of_message = len(bytes_to_send)
        header_bytes = struct.pack("i", length_of_message)
        self.transport.write(header_bytes + bytes_to_send)

    def mark_all_done(self):
        # print(f'[{self.connection_counter}] mark_all_done')

        send_this = CloseConnectionMessage(self.task_id)
        bytes_to_send = pickle.dumps(send_this)
        self.send_with_header(bytes_to_send)

    def connection_lost(self, exc):
        self.timeline.mark_event(f'TCP: Connection to server lost')
        print(f'[{os.getpid()}] [task_id: {self.task_id}] - Child process for test runner is about to exit')
        self.on_con_lost.set_result(True)

    def error_received(self, exc):
        print('Error received:', exc)




import asyncio
import logging.config
from pathlib import Path
import yaml
from aiohttp import web

from pycrunch.session import config
from pycrunch.execution_watchdog.connection_watchdog import connection_watchdog

package_directory = Path(__file__).parent
print(package_directory)
engine_directory = package_directory.parent
config.set_engine_directory(engine_directory)
configuration_yaml_ = package_directory.joinpath('log_configuration.yaml')
print(configuration_yaml_)
with open(configuration_yaml_, 'r') as f:
    logging.config.dictConfig(yaml.safe_load(f.read()))


import sys

if sys.platform == 'win32':
    policy = asyncio.get_event_loop_policy()
    policy._loop_factory = asyncio.ProactorEventLoop


def run():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, help="Port number to listen")
    args = parser.parse_args()
    port = 5000
    if args.port:
        port = args.port
    print(f'PyCrunch port will be {port}')
    use_reloader = not True
    from pycrunch.session.state import engine
    engine.prepare_runtime_configuration_if_necessary()

    from pycrunch.api import shared
    from pycrunch.api.socket_handlers import attach_message_handlers_to_sio

    app = web.Application()

    sio = shared.sio()

    attach_message_handlers_to_sio(sio)
    sio.attach(app)

    if config.enable_web_ui:
        # This will enable PyCrunch web interface
        print(f'PyCrunch Web-UI at http://0.0.0.0:{port}/ui/')
        print(f'                or http://127.0.0.1:{port}/ui/')
        from . import web_ui

        web_ui.enable_for_aiohttp(app, package_directory)
    else:
        print(f'PyCrunch Web-UI is disabled. ')
        print('    To enable it back, please set `engine->enable-web-ui` in `.pycrunch-config.yaml` to true')

    loop = asyncio.get_event_loop()
    task = loop.create_task(connection_watchdog.watch_client_connection_loop())
    loop.set_debug(config.enable_asyncio_debug)

    from pycrunch.compatibility.aiohttp_shim import aiohttp_init_parameters

    additional_kw = aiohttp_init_parameters()

    web.run_app(app, port=port, host='0.0.0.0', shutdown_timeout=1, **additional_kw)


if __name__ == '__main__':
    run()



import inspect
import logging
import sys

from pycrunch.plugins.pytest_support.cleanup_contextmanager import ModuleCleanup
from pycrunch.session import config
from pycrunch.session.file_map import test_map

logger = logging.getLogger(__name__)


class DiscoveredTest:
    def __init__(self, name, filename, module):
        self.filename = filename
        self.module = module
        # name of the test
        self.name = name
        self.fqn = module + ':' + name

class TestSet:
    def __init__(self):
        self.files = set()
        self.tests = []

    def add_module(self, tests_in_module):
        self.files.add(tests_in_module.filename)
        for test in tests_in_module.tests_found:
            self.tests.append(DiscoveredTest(test, tests_in_module.filename, tests_in_module.module))


class TestsInModule:
    def __init__(self, filename, tests_found, module):
        # list of raw strings
        self.tests_found = tests_found
        self.filename = filename
        self.module = module




class SimpleTestDiscovery:

    def __init__(self, root_directory=None, configuration=None):
        self.root_directory = root_directory
        self.configuration = configuration if configuration is not None else config

    def find_tests_in_folder(self, folder, search_only_in = None):

        import glob, importlib, os, pathlib, sys

        if not self.root_directory:
            MODULE_DIR = folder
        else:
            MODULE_DIR = self.root_directory

        logger.debug(sys.path)
        logger.debug(f'MODULE_DIR {MODULE_DIR}')
        logger.debug(f'Discovering tests in folder {folder}')

        if not MODULE_DIR in sys.path:
            sys.path.insert(0, MODULE_DIR)
            logger.debug(f'after append')
            logger.debug(sys.path)
        # The directory containing your modules needs to be on the search path.

        # Get the stem names (file name, without directory and '.py') of any
        # python files in your directory, load each module by name and run
        # the required function.

        parent_path = pathlib.Path(MODULE_DIR)
        folder_path = pathlib.Path(folder)
        # for x in folder_path.glob('**/*.py'):
        #     print('-')
        #     print(module_name_full)

        # py_files = glob.glob(os.path.join(folder, '*.py'))
        py_files = folder_path.glob('**/*.py')
        test_set = TestSet()
        from os import environ
        for env_name, env_value in config.environment_vars.items():
            environ[env_name] = env_value

        self.configuration.prepare_django()
        with ModuleCleanup() as cleanup:

            for py_file in py_files:

                if search_only_in:
                    if str(py_file) not in search_only_in:
                        continue


                # print(py_file)
                current_file_path = py_file.relative_to(parent_path)
                if self.is_excluded_via_configuration(current_file_path):
                    continue

                module_name = self.compute_module_name_from_path(current_file_path)
                # module_name = pathlib.Path(py_file).stem
                if not self.is_module_with_tests(module_name):
                    continue
                try:
                    logger.warning('importing ' + module_name)
                    module = importlib.import_module(module_name)
                    tests_found = self.find_tests_in_module(module)
                except Exception as ex:
                    logger.error(f'Failed to load `{current_file_path}`')
                    logger.exception(f'importing {module_name} failed with exception: ' + str(ex))
                    continue
                # execute as following
                # method_to_call = getattr(module, 'test_1')

                filename = str(py_file)

                test_map.did_found_tests_in_file(filename, tests_found, module_name)
                test_set.add_module(TestsInModule(filename, tests_found, module_name))

                logger.warning(f'tests found: {tests_found}')

        # todo should we mutate global state or return value?
        return test_set

    def compute_module_name_from_path(self, current_file_path):
        if len(current_file_path.parts) > 1:
            module_name = str.join('.', current_file_path.parts[:-1]) + '.' + current_file_path.stem
        else:
            module_name = current_file_path.stem
        return module_name

    def is_excluded_via_configuration(self, current_file_path):
        s = str(current_file_path)
        x = False
        if s.startswith(self.configuration.discovery_exclusions) or s.endswith(self.configuration.discovery_exclusions):
            x = True
        return x

    def find_tests_in_module(self, module):
        all_variables = dir(module)
        found_methods = []
        for v in all_variables:
            function_or_variable_or_class = getattr(module, v)

            if self.is_subclass_of_unittest(function_or_variable_or_class):
                # print(f'{v} : {function_or_variable_or_class} issubclass of unittest')
                attr = getattr(function_or_variable_or_class, "__test__", True)
                # print(attr)
                names = self.get_test_case_names_from_class(function_or_variable_or_class)
                names = map(lambda _: v + '::' + _, names)
                found_methods.extend(names)
                continue

            if self.looks_like_test_name(v):
                function_or_variable_or_class = getattr(module, v)
                if callable(function_or_variable_or_class):
                    found_methods.append(v)
        return found_methods

    def is_subclass_of_unittest(self, function_or_variable_or_class):
        is_class = inspect.isclass(function_or_variable_or_class)
        if not is_class:
            return False

        may_be_not_loaded = sys.modules.get("unittest")
        is_unit_test_sub = may_be_not_loaded and issubclass(function_or_variable_or_class, may_be_not_loaded.TestCase)
        return is_unit_test_sub

    def is_module_with_tests(self, module_name):
        module_short_name = module_name.split('.')[-1]
        return module_short_name.startswith(('test_', 'tests_')) or module_short_name.endswith(('_test','tests', '_tests'))

    def looks_like_test_name(self, v):
        return v.startswith('test_') or v.endswith('_test')

    def get_test_case_names_from_class(self, test_case_class):
        # implementation from standard unitest library
        def is_test_method_in_class(attrname, test_case_class_2=test_case_class):
            startswith = self.looks_like_test_name(attrname)
            can_call = callable(getattr(test_case_class_2, attrname))
            return startswith and can_call

        return list(filter(is_test_method_in_class, dir(test_case_class)))

 # for basecls in inspect.getmro(self.obj.__class__):
 #            dicts.append(basecls.__dict__)


import time
from pprint import pprint
from time import sleep
from unittest import mock

import pytest

from pycrunch.introspection import timings
from pycrunch.introspection.history import execution_history
from pycrunch.introspection.timings import Timeline


def test_timesheet_duration_single_interval():
    sut = Timeline('test 1')
    with use_clock_mock() as clock_mock:
        clock_mock.now.return_value = 1.0
        sut.start()
        clock_mock.now.return_value = 2.0
        sut.stop()
        assert sut.duration() == 1.0
        print_duration(sut)


def test_complex_nesting():
    sut = Timeline('complex')
    # 1 - root
    #   2 ... 4.5 nested_1
    #       3.1 ... 4.1 double_nested
    #  5 end root
    sut.start()

    sut.begin_nested_interval('nested_1')

    sut.begin_nested_interval('a')
    sut.end_nested_interval()

    sut.begin_nested_interval('b')
    sut.end_nested_interval()

    sut.begin_nested_interval('c')
    sut.end_nested_interval()

    sut.end_nested_interval()

    sut.begin_nested_interval('1')
    sut.end_nested_interval()

    sut.begin_nested_interval('2')

    sut.begin_nested_interval('2.1')
    sut.end_nested_interval()

    sut.begin_nested_interval('2.2')

    sut.begin_nested_interval('2.2.1')
    sut.end_nested_interval()

    sut.end_nested_interval()

    sut.end_nested_interval()

    sut.begin_nested_interval('3')
    sut.end_nested_interval()


    sut.stop()
    sut.to_console()

def test_serialization():
    sut = Timeline('complex')
    # 1 - root
    #   2 ... 4.5 nested_1
    #       3.1 ... 4.1 double_nested
    #  5 end root
    sut.start()
    # assert 1 == 2
    sut.begin_nested_interval('nested_1')

    sut.begin_nested_interval('a')
    sut.end_nested_interval()

    sut.begin_nested_interval('b')
    sut.end_nested_interval()

    sut.begin_nested_interval('c')
    sut.end_nested_interval()

    sut.end_nested_interval()

    sut.begin_nested_interval('1')
    sut.end_nested_interval()

    sut.begin_nested_interval('2')

    sut.begin_nested_interval('2.1')
    sut.end_nested_interval()

    sut.begin_nested_interval('2.2')

    sut.begin_nested_interval('2.2.1')
    sut.end_nested_interval()


    sut.end_nested_interval()

    sut.end_nested_interval()
    sut.mark_event('aaa')
    sut.begin_nested_interval('3')
    sut.end_nested_interval()

    sut.stop()

    execution_history.save(sut)
    pprint(execution_history.to_json())

def test_marker_inside_interval():
    sut = Timeline('test 1')
    with use_clock_mock() as clock_mock:
        clock_mock.now.return_value = 1.0
        sut.start()
        clock_mock.now.return_value = 2.0
        sut.begin_nested_interval('nested')
        clock_mock.now.return_value = 3.0
        sut.mark_event('event_1')
        clock_mock.now.return_value = 4.0
        sut.end_nested_interval()
        sut.mark_event('outer')
        clock_mock.now.return_value = 5.0
        sut.stop()
        sut.to_console()

    assert len(sut.root.events) == 1
    assert sut.root.events[0].name == 'outer'
    assert sut.root.events[0].timestamp == 4.0
    assert sut.root.intervals[0].events[0].timestamp == 3.0
    assert sut.root.intervals[0].events[0].name == 'event_1'

def test_double_nesting():
    sut = Timeline('test_double_nest')
    with use_clock_mock() as clock_mock:
        # 1 - root
        #   2 ... 4.5 nested_1
        #       3.1 ... 4.1 double_nested
        #  5 end root
        clock_mock.now.return_value = 1.0
        sut.start()
        clock_mock.now.return_value = 2.0
        sut.begin_nested_interval('nested_1')
        clock_mock.now.return_value = 3.1
        sut.begin_nested_interval('double_nested')
        clock_mock.now.return_value = 4.1

        sut.end_nested_interval()
        clock_mock.now.return_value = 4.5
        sut.end_nested_interval()

        clock_mock.now.return_value = 5
        sut.stop()
        sut.to_console()
        assert len(sut.root.intervals) == 1
        assert sut.root.duration() == 4
        nested_1 = sut.root.intervals[0]
        assert nested_1.duration() == 2.5
        double_nested = nested_1.intervals[0]
        assert double_nested.duration() == pytest.approx(1)

        print_duration(sut)


def test_nested_interval():
    sut = Timeline('test_intervals')
    with use_clock_mock() as clock_mock:
        clock_mock.now.return_value = 1.0
        sut.start()
        clock_mock.now.return_value = 2.0
        sut.begin_nested_interval('nested')
        clock_mock.now.return_value = 12.0

        sut.end_nested_interval()
        clock_mock.now.return_value = 13.0

        sut.stop()
        assert len(sut.root.intervals) == 1
        assert sut.root.duration() == 12.0
        sub_interval = sut.root.intervals[0]
        assert sub_interval.duration() == 10.0
        print_duration(sut)


def test_multiple_intervals():
    sut = Timeline('test_intervals')
    with use_clock_mock() as clock_mock:
        # 1 - root
        #   2 ... 3 nested_1
        #   3.1 ... 4.1 nested_2
        #  5 end root
        clock_mock.now.return_value = 1.0
        sut.start()
        clock_mock.now.return_value = 2.0
        sut.begin_nested_interval('nested_1')
        clock_mock.now.return_value = 3.0
        sut.end_nested_interval()
        clock_mock.now.return_value = 3.1
        sut.begin_nested_interval('nested_2')
        clock_mock.now.return_value = 4.6
        sut.end_nested_interval()

        clock_mock.now.return_value = 5
        sut.stop()

        assert len(sut.root.intervals) == 2
        assert sut.root.duration() == 4
        nested_interval_1 = sut.root.intervals[0]
        assert nested_interval_1.duration() == 1
        nested_interval_2 = sut.root.intervals[1]
        assert nested_interval_2.duration() == pytest.approx(1.5)
        print_duration(sut)



def use_clock_mock():
    return mock.patch('pycrunch.introspection.timings.clock')


def print_duration(sut):
    print(f'start {sut.root.started_at}')
    print(f'end {sut.root.stopped_at}')
    print(f'duration {sut.root.duration()}')
    # dump(sut)


def dump(obj):
  for attr in dir(obj):
    print("obj.%s = %r" % (attr, getattr(obj, attr)))


import logging
from collections import defaultdict

logger = logging.getLogger(__name__)


class TestMap:
    def __init__(self):
        # filename -> test_fqn[]
        self.map = defaultdict(set)

    def did_found_tests_in_file(self, filename, test_names, module_name):
        if len(test_names) <= 0:
            return

        new_list = set()
        for test in test_names:
            new_list.add(module_name + ':' + test)

        self.map[filename] = new_list

    def get_immutable_tests_for_file(self, filename):
        return list(self.map.get(filename, set()))

    def file_did_removed(self, filename):
        if filename not in self.map:
            logger.info(f'file_did_removed could not find  {filename}. Probably it was removed earlier.')
            return
        self.map.pop(filename)

    def test_exist(self, filename, fqn):
        if fqn in self.map[filename]:
            return True

        return False


test_map = TestMap()


import unittest


class MyClass(unittest.TestCase):
    def test_method1(self):
        assert 0 == 0 # fail for demo purposes

    def test_method2(self):
        assert 1 == 1   # fail for demo purposes


class TestForDummies(unittest.TestCase):
    def test_method1(self):
        assert 0 == 0 # fail for demo purposes

    def test_method2(self):
        assert 0 == 0   # fail for demo purposes

    def helper_method(self):
        # should not be in tests
        pass


import io
import pickle
import struct

class ProtocolState():
    def __init__(self, message_queue):
        self.message_queue = message_queue
        # what if header split?
        self.header_size = struct.calcsize("i")
        self.header_buffer = None
        self.in_header = False
        self.header_bytes_so_far = 0

        self.in_message = False
        self.message_bytes_so_far = 0
        self.message_bytes_total = 0
        self.buffer = None
        self.current_payload_size = 0
        self.current_payload_read_so_far = 0

    def feed(self, data):
        """
        This method is responsible for processing input datagrams
        Message may be splitted across multiple datagrams

        Here we read size of next packet from header with fixed byte-length
          and accumulating multiple packages until entire message is read

          also one datagram may contain multiple messages split by header
          so this function solves all of these scenarios
        """
        self.current_payload_size = len(data)
        # print(f'!!! GOT DATAGRAM, size: {self.current_payload_size}')

        self.current_payload_read_so_far = 0
        input_stream = io.BytesIO(data)
        while self.current_payload_read_so_far < self.current_payload_size:
            if not self.in_message:
                # print('reading header')
                if not self.header_buffer:
                    # print('creating new header buffer')
                    self.header_buffer = io.BytesIO()

                self.in_header = True
                # how much we can read?
                read_header_amount = self.header_size - self.header_bytes_so_far
                # print(f'read_header_amount {read_header_amount}')
                # print(f'header_bytes_so_far {self.header_bytes_so_far}')
                if self.current_payload_size - self.current_payload_read_so_far < read_header_amount:
                    read_header_amount = self.current_payload_size - self.current_payload_read_so_far

                message_header_bytes = input_stream.read(read_header_amount)
                self.header_buffer.write(message_header_bytes)
                self.header_bytes_so_far += read_header_amount
                self.current_payload_read_so_far += read_header_amount
                # print(f'current_payload_read_so_far {self.current_payload_read_so_far}')
                if self.header_bytes_so_far >= self.header_size:
                    getbuffer = self.header_buffer.getbuffer()
                    # print(f'self.header_buffer.len() {len(getbuffer)}')
                    self.message_bytes_total = struct.unpack('i', getbuffer.tobytes())[0]
                    # print(f'!!! Expecting message_bytes_total: {self.message_bytes_total}')
                    self.header_buffer = None
                    self.header_bytes_so_far = 0
                    self.in_header = False
                    self.in_message = True
                else:
                    # wait for other part of the header
                    break

            read_amount = self.message_bytes_total - self.message_bytes_so_far
            if self.message_bytes_total - self.message_bytes_so_far > self.current_payload_size - self.current_payload_read_so_far:
                read_amount = self.current_payload_size - self.current_payload_read_so_far

            if not self.buffer:
                # print('creating new buffer')
                self.buffer = io.BytesIO()

            self.buffer.write(input_stream.read(read_amount))
            self.current_payload_read_so_far += read_amount
            self.message_bytes_so_far += read_amount
            if self.message_bytes_so_far >= self.message_bytes_total:
                message = pickle.loads(self.buffer.getbuffer())
                # print(f'message read to the end - {message.kind}')

                self.buffer = None
                self.message_bytes_so_far = 0
                self.in_message = False

                self.message_queue.put_nowait(message)

            if self.current_payload_read_so_far >= self.current_payload_size:
                pass
        # print('read to the end')


import asyncio
import logging
from asyncio import get_event_loop, sleep

from pycrunch.introspection.clock import clock
from pycrunch.pipeline import execution_pipeline
from pycrunch.pipeline.run_test_task import RemoteDebugParams, RunTestTask

logger = logging.getLogger(__name__)


class RunDebouncer:
    def __init__(self, debounce_delay: 'float'):
        self.run_timer = None  # type: 'function'
        now = clock.now()
        self.ts_created = now
        self.ts_target = now + debounce_delay
        self.ts_target = None
        self.run_pending = False
        self.dirty_tests = []
        self.debounce_delay = debounce_delay
        self._loop = get_event_loop()

    def reset_deadline(self, multiplier=1.0):
        now = clock.now()
        self.ts_created = now
        delay_multiplier = self.debounce_delay * multiplier
        if delay_multiplier > 0.85:
            delay_multiplier = 0.85
        self.ts_target = now + delay_multiplier  # (s)

    def delay_a_little(self):
        now = clock.now()
        inside_debounce_interval = self.ts_created < now < self.ts_target
        # print(f' {self.ts_created} < {now} < {self.ts_target}')
        # print(f' === {inside_debounce_interval}')
        if inside_debounce_interval:

            self.ts_target += self.debounce_delay * 1

            if self.ts_target - now > 1.0:
                self.ts_target = 1.0
            if self.ts_target - now < 0:
                self.ts_target = 0.05

            logger.debug(
                f' delay_a_little: updated ts_target, to run in {round(self.ts_target - now, 4)} secs'
            )

            if self.run_pending and self.run_timer:
                logger.debug(
                    f'run_pending, cancelling...'
                )
                copied_ref = self.run_timer
                copied_ref.cancel()
                logger.debug(
                    f' Cancelled pending run. Waiting for more tests to run at once.'
                )
                self.run_timer = None
                self.force_schedule_run()


    def add_tests(self, tests: "List[TestState]"):
        # until we have tests, wait...
        logger.info(f'add_tests: {len(tests)} ')
        if len(self.dirty_tests) <= 0:
            self.reset_deadline(execution_pipeline.tasks_in_queue())
        if len(self.dirty_tests) > 0:
            # we have something in queue, lets wait even more.
            # if there is nothing in queue - intention to run it immediately without waiting
            self.delay_a_little()
        self.dirty_tests.extend(tests)
        logger.debug('2: ran to the end')

    async def schedule_run(self):
        if self.run_pending:
            logger.debug('run_pending already. skip')
            return

        self.run_pending = True
        self.force_schedule_run()

    def force_schedule_run(self):
        self.run_timer = self._loop.create_task(self.execute_with_delay())

    async def execute_with_delay(self):
        target_at_enter = self.ts_target
        try:
            delay = target_at_enter - clock.now()
            if delay < self.debounce_delay:
                delay = self.debounce_delay
            logger.debug(f'execute_with_delay->waiting for {round(delay, 4)} seconds')
            while clock.now() < target_at_enter:
                await sleep(0.01, loop=self._loop)
        except asyncio.CancelledError as e:
            logger.debug(f'execute_with_delay: cancelled to aggregate more tasks')
            return
        except Exception as e:
            logger.exception(f'exc', exc_info=e)
            return

        # after sleep, it may not be a good idea to cancel; use events?
        copy_of_list = list(self.dirty_tests)
        logger.info(
            f'execute_with_delay: Total {len(copy_of_list)} tests will run with delay {round(delay, 4)} seconds',
        )

        execution_pipeline.add_task(
            # While this is running - append only, do not issue another tasks
            RunTestTask(copy_of_list, RemoteDebugParams.disabled())
        )
        self.dirty_tests = []
        self.run_pending = False
        self.reset_deadline()
        self.run_timer = None



import multiprocessing
import subprocess
import sys
from pprint import pprint
from threading import Thread


def nottest_single_multiprocess():
    from multiprocessing.connection import Listener
    address = ('localhost', 6002)  # family is deduced to be 'AF_INET'
    listener = Listener(address, authkey=b'secret password')
    data = '{"tests":[{"fqn":"pycrunch.tests.test_modules_cleanup:test_nested","module":"pycrunch.tests.test_modules_cleanup","filename":"/Users/gleb/code/PyCrunch/pycrunch/tests/test_modules_cleanup.py","name":"test_nested","state":"pending"}]}'

    def thread_loop(params=None):
        print('Waiting for connection')
        conn = listener.accept()
        print('connection accepted from', listener.last_accepted)
        conn.send(data)
        while True:
            msg = conn.recv()
            # do something with msg
            if msg == 'close':
                conn.close()
                break
            else:
                print('got msg from client:')
                pprint(msg)

        listener.close()
    t = Thread(target=thread_loop)
    t.daemon = True
    t.start()
    proc = subprocess.check_call(sys.executable +' /Users/gleb/code/PyCrunch/multiprocess_child_main.py', cwd='/Users/gleb/code/PyCrunch', shell=True)
    pprint(proc)
    t.join(50)
    print('Timeout exceed')




# Interval -> contains start, end, nested intervals, and markers (relative to interval)
import os
from datetime import datetime

from pycrunch.introspection.clock import clock
from collections import deque


class Interval:
    def __init__(self):
        self.intervals = []
        self.events = []
        self.name = None
        self.started_at = None
        self.stopped_at = None

    def duration(self):
        return round(self.stopped_at - self.started_at, 3)

    def start(self):
        self.started_at = clock.now()

    def stop(self):
        self.stopped_at = clock.now()

    def begin_nested_interval(self, name):
        interval = Interval()
        interval.start()
        interval.name = name
        self.intervals.append(interval)
        return interval

    def mark_event(self, name, relative_to):
        self.events.append(Marker(name, relative_to))

    def to_console(self, indent=0):
        join = " ".join([" " for x in range(1, indent + 1)])
        print(f'{join}  {self.name} [{self.started_at} ... {self.stopped_at}] ({self.duration():.3f} seconds)')
        for evt in self.events:
            print(f'    {join}  [event] [{evt.relative_timestamp()}] {evt.name} at {evt.timestamp} (in {self.name}) ')

        for interval in self.intervals:
            interval.to_console(indent=indent+3)



class Marker:
    def __init__(self, name, relative_to):
        self.relative_to = relative_to
        self.timestamp = clock.now()
        self.name = name

    def relative_timestamp(self):
        # relative to entire Timeline
        return round(self.timestamp - self.relative_to, 3)

class Timeline:
    # todo use interval stacks
    def __init__(self, name):
        self.name = name
        self.root = Interval()
        self.root.name = name
        self.intervals_stack = deque()
        self.intervals_stack.append(self.root)
        self.relative_to = None

    def start(self):
        self.relative_to = clock.now()
        self.root.start()

    def stop(self):
        self.root.stop()

    def duration(self):
        # in seconds
        return self.root.duration()

    def begin_nested_interval(self, name):
        self.intervals_stack.append(self.current_interval().begin_nested_interval(name))

    def current_interval(self):
        return self.intervals_stack[-1]

    def end_nested_interval(self):
        current_interval = self.intervals_stack.pop()
        current_interval.stop()

    def to_console(self):
        self.root.to_console()

    def mark_event(self, event_name):
        # print(f'[{os.getpid()}] {datetime.now().isoformat()} {event_name}')
        self.current_interval().mark_event(event_name, self.relative_to)
        pass




import time

from pycrunch.constants import CONFIG_FILE_NAME
from pycrunch.discovery.strategy import create_test_discovery
from pycrunch.pipeline import execution_pipeline
from pycrunch.pipeline.abstract_task import AbstractTask
from pycrunch.pipeline.config_reload_task import ConfigReloadTask
from pycrunch.pipeline.run_debouncing import RunDebouncer
from pycrunch.session import state
from pycrunch.session.combined_coverage import combined_coverage
from pycrunch.session.file_map import test_map

run_debouncer = RunDebouncer(debounce_delay=0.185)

class FileModifiedNotificationTask(AbstractTask):
    def __init__(self, file, context=None):
        self.file = file
        self.timestamp = time.time()
        self.context = context

    async def run(self):
        if self.file.endswith(CONFIG_FILE_NAME):
            execution_pipeline.add_task(ConfigReloadTask())
            return


        # look out for new tests in changed files
        # clean up zombie tests
        # run impacted tests and newly discovered
        # todo: Do not block event_loop!
        discovery = create_test_discovery()
        old_map = test_map.get_immutable_tests_for_file(self.file)
        possibly_new_tests = discovery.find_tests_in_folder(state.engine.folder, search_only_in=[self.file])
        await state.engine.test_discovery_will_become_available(possibly_new_tests)

        new_map = test_map.get_immutable_tests_for_file(self.file)
        removed_tests = set()
        added_tests = set()
        for t in old_map:
            if t not in new_map:
                removed_tests.add(t)

        for t in new_map:
            if t not in old_map:
                added_tests.add(t)

        execution_plan = set()
        for new_test in possibly_new_tests.tests:
            if new_test.fqn in new_map:
                execution_plan.add(new_test.fqn)

        dependencies = combined_coverage.dependencies
        if dependencies:
            impacted_tests = dependencies[self.file]

            for fqn in impacted_tests:
                if fqn not in removed_tests:
                    execution_plan.add(fqn)
                else:
                    print(f"test {fqn} removed from execution plan")

            if state.config.engine_mode == 'manual':
                print('Manual mode, tests wont run. Consider switching engine mode to auto')
                return

            tests_to_run = state.engine.all_tests.collect_by_fqn(execution_plan)
            dirty_tests = self.consider_engine_mode(tests_to_run)
            run_debouncer.add_tests(dirty_tests)
            await run_debouncer.schedule_run()


    def consider_engine_mode(self, tests_to_run):
        if state.config.engine_mode == 'auto':
            return tests_to_run

        if state.config.engine_mode == 'pinned':
            only_pinned = []
            for test in tests_to_run:
                if test.pinned:
                    only_pinned.append(test)
            return only_pinned
        print('Cannot filter by engine mode.')
        return tests_to_run

# https://stackoverflow.com/questions/45369128/python-multithreading-queue



from unittest.mock import patch

from pycrunch.compatibility.aiohttp_shim import aiohttp_init_parameters


def test_aiohttp_init_parameters_3_7():
    with patch('pycrunch.compatibility.aiohttp_shim.aiohttp') as mock:
        mock.__version__ = '3.7'
        actual = aiohttp_init_parameters()
        assert 'loop' not in actual


def test_aiohttp_init_has_loop_since_3_8():
    with patch('pycrunch.compatibility.aiohttp_shim.aiohttp') as mock:
        mock.__version__ = '3.8'
        actual = aiohttp_init_parameters()
        assert 'loop' in actual



from datetime import date, datetime
from unittest.mock import MagicMock

from pycrunch.insights.variables_inspection import InsightTimeline, RecordedVariable, trace


def test_should_adjust_to_clock_start():
    clock = MagicMock()
    now_mock = MagicMock()
    now_mock.side_effect = (100, 101, 110)

    # 100 -- 101 ----------------- 110
    # events should be in timeline time:
    # 1 ----------------- 10
    clock.now = now_mock
    x = InsightTimeline(clock)
    x.start()
    x.record(x=1)
    x.record(y=2)
    variables = x.variables
    assert variables[0].timestamp == 1
    assert variables[1].timestamp == 10

def test_non_traceble_variables_should_cast_to_str():
    sut = RecordedVariable('name', datetime.now(), -1)
    actual = sut.as_json()
    assert type(actual['value']) == str



from time import sleep
from unittest.mock import MagicMock

from pycrunch.insights.variables_inspection import InsightTimeline, trace


def test_random_stuff():
    for x in range(10):
        # sleep(0.5)
        trace(x=x)




import io
from pathlib import Path

# from flask import jsonify, Response
# from flask_socketio import send

from pycrunch import runner
from pycrunch.api.shared import file_watcher
from pycrunch.pipeline import execution_pipeline
from pycrunch.pipeline.run_test_task import RunTestTask
from pycrunch.session import config
from pycrunch.session.state import engine
from pycrunch.shared.models import all_tests

import logging

logger = logging.getLogger(__name__)


# @pycrunch_api.route("/")
# def hello():
#     return "Nothing here"


# @pycrunch_api.route("/discover")
# def discover_tests():
#     engine.will_start_test_discovery()
#     return jsonify(dict(ack=True))

#
# @pycrunch_api.route("/diagnostics")
# def queue_diagnostics():
#     engine.will_start_diagnostics_collection()
#     return jsonify(dict(ack=True))


# @pycrunch_api.route("/timings")
# def queue_timings():
#     engine.will_send_timings()
#     return jsonify(dict(ack=True))


# @pycrunch_api.route("/pin-tests", methods=['POST'])
# def pin_tests():
#     # fqns = array of strings[]
#     fqns = request.json.get('fqns')
#     engine.tests_will_pin(fqns)
#     return jsonify(dict(ack=True))


# @pycrunch_api.route("/engine-mode", methods=['POST'])
# def engine_mode():
#     new_mode = request.json.get('mode')
#     engine.engine_mode_will_change(new_mode)
#     return jsonify(dict(ack=True))

# @pycrunch_api.route("/unpin-tests", methods=['POST'])
# def unpin_tests():
#     # fqns = array of strings[]
#     fqns = request.json.get('fqns')
#     engine.tests_will_unpin(fqns)
#     return jsonify(dict(ack=True))
#
# @pycrunch_api.route("/run-tests", methods=['POST'])
# def run_tests():
#
#     tests = request.json.get('tests')
#     logger.info('Running tests...')
#     logger.info(tests)
#     fqns = set()
#     for test in tests:
#         fqns.add(test['fqn'])
#
#     tests_to_run = all_tests.collect_by_fqn(fqns)
#
#     execution_pipeline.add_task(RunTestTask(tests_to_run))
#     return jsonify(tests)



# @pycrunch_api.route("/file", methods=['GET'])
# def download_file():
#     filename = request.args.get('file')
#     logger.debug('download_file ' + filename)
#     target_file = config.path_mapping.map_local_to_remote(filename)
#     my_file = io.FileIO(target_file, 'r')
#     content = my_file.read()
#     return Response(content, mimetype='application/x-python-code')




import uuid

from pycrunch.scheduling.scheduler import TestRunScheduler
from pycrunch.shared.primitives import TestMetadata

def generate_test_fake(name):
    filename = str(uuid.uuid4())
    test_name = str(name)
    return TestMetadata(filename, test_name, str(uuid.uuid4()), test_name, 'unknown')


def generate_tests(desired_amount):
    result = []
    for n in range(0, desired_amount):
        result.append(generate_test_fake(n))

    return result


def test_scheduled_single_test():
    tests = generate_tests(1)
    sut = create_sut()
    actual = sut.schedule_into_tasks(tests)
    # actual is list of tasks
    assert len(actual) == 1
    first_task = actual[0]
    assert len(first_task.tests) == 1

def test_up_to_five_tests_should_be_in_single_task():
    tests = generate_tests(5)
    sut = create_sut()
    tasks = sut.schedule_into_tasks(tests)
    assert len(tasks) == 1
    first_task = tasks[0]
    assert len(first_task.tests) == 5


def test_six_tests_should_be_in_two_tasks():
    tests = generate_tests(6)
    sut = create_sut()
    tasks = sut.schedule_into_tasks(tests)
    assert len(tasks) == 2
    assert len(tasks[0].tests) == 3
    assert len(tasks[1].tests) == 3

def test_seven_tests_should_be_in_two_tasks():
    tests = generate_tests(7)
    sut = create_sut()
    tasks = sut.schedule_into_tasks(tests)
    assert len(tasks) == 2
    first_task = tasks[0]
    second_task = tasks[1]
    assert len(first_task.tests) == 4
    assert len(second_task.tests) == 3


def test_14_should_be_in_tree_tasks():
    tests = generate_tests(14)
    sut = create_sut()
    tasks = sut.schedule_into_tasks(tests)
    assert len(tasks) == 3
    assert len(tasks[0].tests) == 5
    assert len(tasks[1].tests) == 5
    assert len(tasks[2].tests) == 4


def test_twenty_should_be_in_four_tasks():
    tests = generate_tests(20)
    sut = create_sut()
    tasks = sut.schedule_into_tasks(tests)
    assert len(tasks) == 4
    assert len(tasks[0].tests) == 5
    assert len(tasks[1].tests) == 5
    assert len(tasks[2].tests) == 5
    assert len(tasks[3].tests) == 5

def test_51_should_be_in_four_tasks():
    tests = generate_tests(51)
    sut = create_sut()
    tasks = sut.schedule_into_tasks(tests)
    assert len(tasks) == 4
    assert len(tasks[0].tests) == 13
    assert len(tasks[1].tests) == 13
    assert len(tasks[2].tests) == 13
    assert len(tasks[3].tests) == 12


def create_sut():
    return TestRunScheduler(cpu_cores=4, threshold=5)



def test_generate_code():
    template =\
"""

def test_{number}():
    print('number is {number}')

"""
    # useful for testing a lot of tests
    for x in range(1, 1000):
        print(template.replace('{number}', str(x)))



from unittest import mock
from unittest.mock import mock_open, patch, Mock

import pytest

from pycrunch import session


def test_pytest_engine_by_default():
    sut = create_sut()
    assert sut.runtime_engine == 'pytest'


def test_deep_inheritance_is_off_by_default():
    sut = create_sut()
    assert not sut.deep_inheritance


def test_deep_inheritance_is_on_in_config():
    read_data = '''
       engine:
         deep-inheritance: true
       '''
    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert sut.deep_inheritance


def test_can_change_to_simple_engine():
    sut = create_sut()
    sut.runtime_engine_will_change('simple')
    assert sut.runtime_engine == 'simple'


def test_can_change_to_django_engine():
    sut = create_sut()
    sut.runtime_engine_will_change('django')
    assert sut.runtime_engine == 'django'


def create_sut():
    configuration = session.configuration.Configuration()
    configuration.watch_for_config_changes = Mock()
    return configuration


def test_should_call_auto_config():
    with patch(
        'pycrunch.session.configuration.AutoConfiguration.ensure_configuration_exist'
    ) as auto:
        sut = create_sut()
        sut.load_runtime_configuration()
        auto.assert_called_once()


def test_non_supported_engine_throws():
    import pytest

    sut = create_sut()
    with pytest.raises(Exception):
        sut.runtime_engine_will_change('not-supported')


read_data = '''
engine:
  runtime: simple
  timeout: 42
discovery:
  exclusions:
   - directory_1
   - directory_2
mode: manual

'''


def test_exclusion_list():
    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert 'directory_1' in sut.discovery_exclusions
        assert 'directory_2' in sut.discovery_exclusions
        assert 'directory_3' not in sut.discovery_exclusions


def test_coverage_excluded():
    read_data = '''coverage-exclusions:
   - 'my/dir/_module'
   - 'my/dir2/_module2'
   - 'file.py'
    '''
    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert '*my/dir/_module*' in sut.coverage_exclusions
        assert '*my/dir2/_module2*' in sut.coverage_exclusions
        assert '*file.py' in sut.coverage_exclusions


def test_engine_mode_auto_by_default():
    sut = create_sut()
    assert sut.engine_mode == 'auto'


def test_unsupported_engine_mode_raises_exception():
    sut = create_sut()
    with pytest.raises(Exception):
        sut.runtime_mode_will_change('no-way')


def test_runtime_engine_from_config():
    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert sut.runtime_engine == 'simple'


def test_timeout_is_1min_by_default():
    sut = create_sut()
    assert sut.execution_timeout_in_seconds == 60


def test_timeout_is_none_when_zero():
    # this will force task to wait forever
    sut = create_sut()
    sut.execution_timeout_in_seconds = 0
    assert sut.get_execution_timeout() is None


def test_timeout_is_taken_from_config():
    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()

        assert sut.execution_timeout_in_seconds == 42


def test_environment_vars():
    read_data = '''
    discovery:
      exclusions:
       - directory_1
       - directory_2
    engine:
      runtime: simple
      mode: manual
    env:
      DJANGO_SETTINGS: insuredportal.settings
      ENV2: test
    '''

    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert 'insuredportal.settings' == sut.environment_vars.get('DJANGO_SETTINGS')
        assert 'test' == sut.environment_vars.get('ENV2')
        assert sut.environment_vars.get('NOT') == None


def test_path_map_config_read():
    read_data = '''
    discovery:
      exclusions:
       - directory_1
       - directory_2
    engine:
      runtime: simple
      mode: manual
    env:
      DJANGO_SETTINGS: insuredportal.settings
      ENV2: test
    path-mapping:
      /code: /localfs/easyproject 
    '''

    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert '/code' == sut.path_mapping.path_in_container
        assert '/localfs/easyproject' == sut.path_mapping.path_on_local_ide


def test_cpu_cores_considered():
    read_data = '''
    discovery:
      exclusions:
       - directory_1
       - directory_2
    engine:
      runtime: simple
      cpu-cores: 42
    '''

    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert sut.cpu_cores == 42


def test_cpu_cores_by_default_no_more_than_4():
    with mock.patch(
        'pycrunch.session.configuration.multiprocessing'
    ) as multiprocessing_mock:
        # lets just use 4 by default to not burn machine
        multiprocessing_mock.cpu_count.return_value = 16
        sut = create_sut()
        assert sut.cpu_cores == 4


def test_two_cores_use_1_by_default():
    with mock.patch(
        'pycrunch.session.configuration.multiprocessing'
    ) as multiprocessing_mock:
        multiprocessing_mock.cpu_count.return_value = 2
        sut = create_sut()
        assert sut.cpu_cores == 1


def test_one_core_use_1_by_default():
    with mock.patch(
        'pycrunch.session.configuration.multiprocessing'
    ) as multiprocessing_mock:
        multiprocessing_mock.cpu_count.return_value = 1
        sut = create_sut()
        assert sut.cpu_cores == 1


def test_from_3_to_7_cores_use_half_of_them_by_default():
    with mock.patch(
        'pycrunch.session.configuration.multiprocessing'
    ) as multiprocessing_mock:
        # lets just use 4 by default to not burn machine
        multiprocessing_mock.cpu_count.return_value = 6
        sut = create_sut()
        assert sut.cpu_cores == 3


def test_multiprocessing_threshold_considered():
    read_data = '''
    engine:
      runtime: simple
      multiprocessing-threshold: 1
    '''

    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert sut.multiprocessing_threshold == 1


def test_multiprocessing_threshold_by_default_5():
    sut = create_sut()
    # minimum number of tests to schedule per core
    assert sut.multiprocessing_threshold == 5


def test_load_pytest_plugins_false_by_default():
    sut = create_sut()
    assert sut.load_pytest_plugins == False


def test_load_pytest_plugins():
    read_data = '''
engine:
  runtime: simple
  load-pytest-plugins: true
discovery:
  exclusions:
   - directory_1
   - directory_2
mode: manual

'''
    with mock.patch('io.open', mock_open(read_data=read_data)) as x:
        sut = create_sut()
        sut.load_runtime_configuration()
        assert sut.load_pytest_plugins == True



from asyncio import get_event_loop
from typing import Any, Dict

import aiohttp
from pycrunch.compatibility.version_utils import parse_version_string


def aiohttp_init_parameters() -> Dict[str, Any]:

    # aiohttp since 3.8 creates new event loop on startup.
    #   https://github.com/aio-libs/aiohttp/pull/5572
    # aiohttp <= 3.7 does not have `loop` argument, we do not want to pass it

    result_arguments = {}
    aio_version = parse_version_string(aiohttp.__version__)
    if aio_version >= (3, 8, 0):
        result_arguments['loop'] = get_event_loop()

    return result_arguments



import math
from typing import List

from pycrunch.scheduling.sheduled_task import TestRunPlan

import logging

logger = logging.getLogger(__name__)


class TestRunScheduler():
    # Do not split up to 5 tasks
    threshold = 5

    def __init__(self, cpu_cores, threshold):
        self.cpu_cores = cpu_cores
        self.threshold = threshold

    def schedule_into_tasks(self, tests) -> List[TestRunPlan]:
        list_of_tasks = []
        total_tests_to_run = len(tests)
        logger.info(f'total_tests_to_run {total_tests_to_run}')
        if total_tests_to_run <= self.threshold:
            list_of_tasks.append(TestRunPlan(tests))
            return list_of_tasks

        avarage_per_core = total_tests_to_run / self.cpu_cores
        total_cores_to_use = 1
        tests_can_be_scheduled = 0
        if avarage_per_core < self.threshold:
            for x in range(1, self.cpu_cores):
                total_cores_to_use = x
                tests_can_be_scheduled += self.threshold
                if tests_can_be_scheduled > total_tests_to_run:
                    break
            logger.debug(f'average_per_core: {avarage_per_core}')
        else:
            total_cores_to_use = self.cpu_cores
        logger.debug(f'tests_can_be_scheduled {tests_can_be_scheduled}')
        logger.info(f'total_cores_to_use {total_cores_to_use}')
        tests_so_far = 0
        tests_scheduled_per_core_with_decimal = total_tests_to_run / total_cores_to_use
        logger.debug(f'tests_scheduled_per_core_with_decimal {tests_scheduled_per_core_with_decimal}')
        tests_to_run_on_single_core = math.ceil(tests_scheduled_per_core_with_decimal)
        logger.info(f'rounded tests_to_run_on_single_core {tests_to_run_on_single_core}')
        for x in range(total_cores_to_use):
            from_index = tests_so_far
            to_index = tests_so_far + tests_to_run_on_single_core
            logger.debug(f'from:to -> {from_index}:{to_index}')
            spliced = tests[from_index:to_index]
            logger.debug(f'core {x}; spliced amount: {len(spliced)}')
            plan = TestRunPlan(spliced)
            list_of_tasks.append(plan)
            tests_so_far += tests_to_run_on_single_core
        return list_of_tasks



class TestMetadata:
    def __init__(self, filename, name, module, fqn, state):
        self.state = state
        self.fqn = fqn
        self.module = module
        self.name = name
        self.filename = filename



import coverage

from pycrunch.api.serializers import CoverageRunForSingleFile


class CoverageAbstraction:
    cov: "coverage.Coverage"

    def __init__(self, disable,
                 coverage_exclusions,  # type: list[str]
                 timeline):
        """

        :type timeline: pycrunch.introspection.timings.Timeline
        :type disable: bool; coverage will be disabled inside debugging session
        """
        self.timeline = timeline
        self.disable = disable
        self.coverage_exclusions = coverage_exclusions
        self.cov = None

    def start(self):
        if self.disable:
            self.timeline.mark_event('Run: Coverage disabled. Do not start it.')
            return

        from . import exclusions

        use_slow_tracer = False
        # user_slow_tracer = True
        coverage_args = self.get_coverage_arguments()

        # todo exclusion list should be configurable
        cov = coverage.Coverage(
            **coverage_args,
            config_file=False,
            timid=use_slow_tracer,
            branch=False,
            omit=exclusions.exclude_list + self.coverage_exclusions,
        )

        # logger.debug('-- before coverage.start')

        # !!!!!
        # debug will not work after cov.start is called!
        # !!!!!
        cov.start()

        # cov.exclude('def')

        # logger.debug('-- after coverage.start')
        self.timeline.mark_event('Run: Coverage started')
        self.cov = cov

    def stop(self):
        if self.disable:
            self.timeline.mark_event('Run: Coverage disabled. Cannot stop.')
            return

        self.cov.stop()
        self.timeline.mark_event('Coverage stopped')

    def get_coverage_arguments(self):
        coverage_args = dict()
        if self.is_coverage_v5_or_greater():
            # dont write on disk `.coverage` files,
            # it is not needed, and will be deadlocked due to concurrent test execution
            coverage_args.update(
                dict(
                    data_file=None
                ))
        return coverage_args

    def is_coverage_v5_or_greater(self):
        version_info = getattr(coverage, 'version_info', None)
        if not isinstance(version_info, tuple) or len(version_info) <= 0:
            return False

        return version_info[0] >= 5

    def parse_all_hit_lines(self):
        """
          Converts data from coverage.py format to internal format

          And plans are to integrate it with PyTrace Coverage
          But then pytrace should be optimized/compiled for multi-platform

          :rtype: typing.List[pycrunch.api.serializers.CoverageRunForSingleFile]
        """

        if self.disable:
            return []

        interim_results = []
        coverage_data = self.cov.get_data()
        for f in coverage_data.measured_files():
            # maybe leave only what we need?
            lines = coverage_data.lines(f)

            # arcs = coverage_data.arcs(f)
            #         * The file name for the module.
            #         * A list of line numbers of executable statements.
            #         * A list of line numbers of excluded statements.
            #         * A list of line numbers of statements not run (missing from
            #           execution).
            #         * A readable formatted string of the missing line numbers.
            # // todo exclude lines hits
            # analysis = cov.analysis2(f)
            interim_results.append(CoverageRunForSingleFile(f, lines))
        return interim_results



# import asyncio
# import pickle
# import time
#
# import pytest
#
# from pycrunch.introspection.timings import Timeline
# from pycrunch.scheduling.client_protocol import EchoClientProtocol
# from pycrunch.scheduling.sheduled_task import TestRunPlan
# from pycrunch.tests.test_scheduling import generate_tests
#
# # async def tcp_echo_client(message):
# #     reader, writer = await asyncio.open_connection('127.0.0.1', 8888)
# #
# #     print('Send: %r' % message)
# #     writer.write(message.encode())
# #
# #     data = await reader.read(100)
# #     print('Received: %r' % data.decode())
# #
# #     print('Close the socket')
# #     writer.close()
#
# @pytest.mark.asyncio
# async def test_connection_is_made():
#     tests = generate_tests(3)
#     plan = []
#     plan.append(TestRunPlan(tests, id='1'))
#     plan.append(TestRunPlan(tests, id='2'))
#     plan.append(TestRunPlan(tests, id='3'))
#
#     sut = ConcurrentMultiprocessTestRunner(plan)
#     loop = asyncio.get_event_loop()
#
#     on_con_lost = loop.create_future()
#     on_con_lost2 = loop.create_future()
#     on_con_lost3 = loop.create_future()
#     message = 'Hello World!'
#     for x in range(10):
#         message += str(x)
#     # await future_x
#     # server_task = loop.create_task(sut.run(plan))
#     # await asyncio.sleep(0.1)
#     t = Timeline('test run fake')
#     t.start()
#     transport, protocol1 = await loop.create_connection(
#         lambda: EchoClientProtocol(on_con_lost, plan[0].id, t),
#         '127.0.0.1', 8888)
#
#     transport2, protocol2 = await loop.create_connection(
#         lambda: EchoClientProtocol(on_con_lost2, plan[1].id, t),
#         '127.0.0.1', 8888)
#
#     transport3, protocol3 = await loop.create_connection(
#         lambda: EchoClientProtocol(on_con_lost3, plan[2].id, t),
#         '127.0.0.1', 8888)
#     # Wait until the protocol signals that the connection
#     # is lost and close the transport.
#     await asyncio.sleep(3)
#     protocol1.mark_all_done()
#     protocol2.mark_all_done()
#     protocol3.mark_all_done()
#     try:
#         await on_con_lost
#         await on_con_lost2
#         await on_con_lost3
#     finally:
#         transport.close()
#         transport2.close()
#         transport3.close()


import importlib.util
import logging
import sys
import traceback

from pycrunch.runner import _abstract_runner
from pycrunch.runner.execution_result import ExecutionResult

logger = logging.getLogger(__name__)


class SimpleTestRunnerEngine(_abstract_runner.Runner):
    """
      This class probably is not used anymore. It was for early prototype.
    """
    def run_test(self, test):
        execution_result = ExecutionResult()
        try:
            logger.debug('before _run_module...')
            spec = importlib.util.spec_from_file_location("fake.name", test.filename)
            logger.debug('  spec_from_file_location -> done; importing module...')

            foo = importlib.util.module_from_spec(spec)

            logger.debug('  module_from_spec -> done; going to exec_module...')
            # logger.warning(f'_run_test->vars {vars(foo)}')
            spec.loader.exec_module(foo)
            method_to_call = getattr(foo, test.name, None)
            # print(vars(foo))
            if method_to_call:
                method_to_call()
                execution_result.run_did_succeed()
            # logger.debug('after exec_module')
        except Exception as e:
            etype, value, current_traceback = sys.exc_info()
            execution_result.record_exception(etype=etype, value=value, current_traceback=current_traceback)
            execution_result.run_did_fail()
            last_call = -1
            traceback.print_exception(etype=etype, value=value, tb=current_traceback, limit=last_call, file=sys.stdout)
            # traceback.print_exc(file=sys.stdout)
            # print(str(e))
            logger.exception('Error while executing _run_test', exc_info=e)

        return execution_result



import logging
from asyncio import Queue

logger = logging.getLogger(__name__)


class ExecutionPipeline:
    def __init__(self):
        self.q = Queue()

    def add_task(self, task: "AbstractTask"):
        logger.debug('Received task in queue')
        self.q.put_nowait(task)

    def tasks_in_queue(self):
        return self.q.qsize()

    async def get_task(self):
        return await self.q.get()

    async def put_raw(self, abstract_task):
        await self.q.put(abstract_task)


execution_pipeline = ExecutionPipeline()



from asyncio import shield

import socketio

from pycrunch.watcher.fs_watcher import FSWatcher
# from pycrunch.watcher.fs_watcher_heavy import FSWatcher

file_watcher = FSWatcher()


# async_mode='threading'
# async_mode='tornado'
# async_mode='eventlet'
async_mode='aiohttp'

# log_ws_internals = True
log_ws_internals = False
sio_instance = None


def sio():
    global sio_instance
    if sio_instance is None:
        sio_instance = socketio.AsyncServer(async_mode=async_mode, cors_allowed_origins='*', logger=log_ws_internals, engineio_logger=log_ws_internals)
    return sio_instance



class ExternalPipe:
    async def push(self, event_type, **kwargs):
        # print(f'ws event: {event_type}')
        # pprint(kwargs)

        await shield(sio().emit('event',
                       dict(
                           event_type=event_type,
                           **kwargs
                       ),
                       namespace='/'))


pipe = ExternalPipe()


import os
import sys


class DiagnosticEngine:

    def summary(self):
        return dict(env=self.get_env(), modules=self.get_modules())

    def get_env(self):
        return dict(os.environ.items())

    def get_modules(self):
        return list(sys.modules)

diagnostic_engine = DiagnosticEngine()


import time

from pycrunch.api import shared
from pycrunch.pipeline.abstract_task import AbstractTask
from pycrunch.session.file_map import test_map


class FileRemovedTask(AbstractTask):
    def __init__(self, file):
        self.file = file
        self.timestamp = time.time()

    async def run(self):
        from pycrunch.session import state
        await shared.pipe.push(event_type='file_modification',
                               modified_file=self.file,
                               ts=self.timestamp,
                               )

        test_map.file_did_removed(self.file)
        from pycrunch.discovery.simple import TestSet
        await state.engine.test_discovery_will_become_available(TestSet())



from collections import OrderedDict

from pycrunch.session import config


class CoverageRunForSingleFile:
    def __init__(self, filename, lines):
        self.lines = lines
        self.filename = filename

    def as_json(self):
        return OrderedDict(filename=config.path_mapping.map_to_local_fs(self.filename), lines_covered=self.lines,)

class CoverageRun:
    def __init__(self, fqn, time_elapsed, test_metadata, execution_result):
        self.test_metadata = test_metadata
        self.time_elapsed = time_elapsed
        self.entry_point = fqn
        self.execution_result = execution_result
        self.percentage_covered = -1
        self.files = []

    def store_files_coverage(self, files):
        """
        :type files: typing.List[CoverageRunForSingleFile]
        """
        self.files = files



    def as_json(self):
        files_ = [x.as_json() for x in self.files]
        return dict(
            entry_point=self.entry_point,
            time_elapsed=round(self.time_elapsed * 1000, 2),
            test_metadata=self.test_metadata,
            files=files_,
            status=self.execution_result.status,
            captured_output=self.execution_result.captured_output,
            variables_state=self.execution_result.state_timeline.as_json(),
        )





def serialize_test_set_state(test_set):
    def serialize_test(test_state):
        """

        :type test_state: pycrunch.shared.models.TestState
        """
        discovered_test = test_state.discovered_test
        execution_result = test_state.execution_result
        return dict(
            fqn=discovered_test.fqn,
            module=discovered_test.module,
            filename=config.path_mapping.map_to_local_fs(discovered_test.filename),
            name=discovered_test.name,
            state=execution_result.status,
            pinned=test_state.pinned,
        )

    return dict(
        tests=[serialize_test(v) for (k, v) in test_set.items()],
        )


from pprint import pprint


class PyTestInterceptionPlugin:
    def __init__(self, tests_to_run):
        self.passed_tests = set()
        self.failed_tests = set()

    def pytest_runtest_setup(item):
        # called for running each test in 'a' directory
        # print("setting up", item)
        pass

    def pytest_runtest_logstart(nodeid, location):
        """
            :param str nodeid: full id of the item
            :param location: a triple of ``(filename, linenum, testname)``
        """
        # print(f"pytest_runtest_logstart {location}", nodeid)
        pass

    def pytest_runtest_logfinish(nodeid, location):
        """ signal the complete finish of running a single test item.

        This hook will be called **after** :func:`pytest_runtest_setup`, :func:`pytest_runtest_call` and
        :func:`pytest_runtest_teardown` hooks.

        :param str nodeid: full id of the item
        :param location: a triple of ``(filename, linenum, testname)``
        """
        # print(f"pytest_runtest_logfinish {location}", nodeid)

    def pytest_exception_interact(node, call, report):
        """called when an exception was raised which can potentially be
        interactively handled.

        This hook is only called if an exception was raised
        that is not an internal exception like ``skip.Exception``.
        """
        # pprint(call)
        # print(f"pytest_exception_interact {call}")

        # pprint(vars(report))
        pass

    def pytest_runtest_logreport(self, report):
        # pprint(vars(report))
        if report.when == 'setup':
            if report.outcome == 'skipped':
                self.passed_tests.add(report.nodeid)

        if report.when == 'teardown':
            pass

        if report.when == 'call':
            if report.outcome == 'skipped':
                self.passed_tests.add(report.nodeid)
                return
            if report.passed:
                self.passed_tests.add(report.nodeid)
            if not report.passed:
                self.failed_tests.add(report.nodeid)

    def guess_run_status(self, test_name):
        if len(self.failed_tests) > 0:
            return False

        if len(self.passed_tests) <= 0:
            return False
        else:
            return True



import io
from pathlib import Path

class AutoConfiguration:
    def __init__(self, configuration_file):
        """
        :type configuration_file: Path
        """
        self.configuration_file = configuration_file

    def ensure_configuration_exist(self):
        if self.configuration_file.exists():
            return
        
        self.create_default_configuration_file()

    def create_default_configuration_file(self):
        default_config_file = '''# documentation https://pycrunch.com/docs/configuration-file
engine:
    runtime: pytest
    enable-web-ui: false
    load-pytest-plugins: false
'''
        with io.open(self.configuration_file, encoding='utf-8', mode='w') as f:
            f.write(default_config_file)

        print(f'Created default configuration file.')



import aiohttp
from aiohttp import web


def enable_for_aiohttp(app, package_directory):
    async def root_handler(request):
        return aiohttp.web.HTTPFound('/ui/index.html')

    async def favicon_handler(request):
        # return aiohttp.web.HTTPFound('/ui/favicon.ico')
        return web.FileResponse(package_directory.joinpath('web-ui','favicon.ico'))

    app.router.add_route('*', '/ui/', root_handler)
    app.router.add_route('*', '/favicon.ico', favicon_handler)
    app.add_routes([web.static('/ui', package_directory.joinpath('web-ui'))])
    app.add_routes([web.static('/js', package_directory.joinpath('web-ui', 'js'))])
    app.add_routes([web.static('/css', package_directory.joinpath('web-ui', 'css'))])
    app.add_routes([web.static('/fonts', package_directory.joinpath('web-ui', 'fonts'))])


py_test_exclusions = [
    '*_pytest/*',
    '*site-packages/*',
    '*.pytest_cache/*',
    '*pytest_support/interception_plugin.py',

]

exclude_list = py_test_exclusions + [
    '*PyCharm.app/Contents/helpers/pydev/pydevd_file_utils.py',
    '*PyCharm.app/Contents/helpers/pydev/_pydevd_bundle/pydevd_comm.py',
    '*PyCharm.app/Contents/helpers/pydev/_pydevd_frame_eval/pydevd_frame_tracing.py',
    '*PyCharm.app/Contents/helpers/pydev/_pydevd_frame_eval/pydevd_modify_bytecode.py',
    '*PyCharm.app/Contents/plugins/python/helpers/pydev/*',
    '*site-packages/werkzeug/serving.py',
    '*runner/test_runner.py',
    '*runner/execution_result.py',
    '*plugins/pytest_support/hot_reload.py',
    '*plugins/pytest_support/pytest_runner_engine.py',
    '*runner/interception.py',
    '*pycrunch/introspection/*',
    '*pycrunch/shared/models.py'

]




from pycrunch.pipeline.abstract_task import AbstractTask
from pycrunch.session import config


class ConfigReloadTask(AbstractTask):
    async def run(self):
        config.load_runtime_configuration()



import sys
from pathlib import Path
from pprint import pprint

import yaml

from pycrunch.plugins.pytest_support.hot_reload import unload_candidates
import logging.config
import logging

logger = logging.getLogger(__name__)


# http://forums.cgsociety.org/t/proper-way-of-reloading-a-python-module-with-new-code-without-having-to-restart-maya/1648174/8

class ModuleCleanup():
    def __init__(self):
        self.modules_before = []

    def __enter__(self):
        self.modules_before = set(sys.modules)
        # self.apply_log_config()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        modules_after = set(sys.modules)
        difference = set(modules_after).difference(self.modules_before)
        modules_for_unload = unload_candidates(difference)
        # pprint(modules_for_unload)
        for m in modules_for_unload:
        # wont work witth django without check?! find out why?
            if m in sys.modules:
                del sys.modules[m]

    def apply_log_config(self):
        parent = Path(__file__).parent.parent.parent
        print(parent)
        configuration_yaml_ = parent.joinpath('log_configuration.yaml')
        print(configuration_yaml_)
        with open(configuration_yaml_, 'r') as f:
            logger.info('aaa')
            logging.config.dictConfig(yaml.safe_load(f.read()))
            logger.info('bbb')



from io import StringIO
import sys
from contextlib import contextmanager

#  from https://stackoverflow.com/questions/17067560/intercept-pythons-print-statement-and-display-in-gui
@contextmanager
def capture_stdout():
    """
    context manager encapsulating a pattern for capturing stdout writes
    and restoring sys.stdout even upon exceptions

    Examples:
    >>> with capture_stdout() as get_value:
    >>>     print("here is a print")
    >>>     captured = get_value()
    >>> print('Gotcha: ' + captured)

    >>> with capture_stdout() as get_value:
    >>>     print("here is a print")
    >>>     raise Exception('oh no!')
    >>> print('Does printing still work?')
    """
    # Redirect sys.stdout
    out = StringIO()
    sys.stdout = out
    # Yield a method clients can use to obtain the value
    try:
        yield out.getvalue
    finally:
        # Restore the normal stdout
        sys.stdout = sys.__stdout__


import asyncio
import logging
import os
import sys
from typing import List, Optional

from pycrunch.networking.server_protocol import TestRunnerServerProtocol
from pycrunch.session import config

logger = logging.getLogger(__name__)


class MultiprocessTestRunner:
    def __init__(self, timeout: Optional[float], timeline, test_run_scheduler, remote_debug_params: "RemoteDebugParams"):
        self.client_connections: List[TestRunnerServerProtocol] = []
        self.completion_futures = []
        self.timeline = timeline
        self.timeout = timeout
        self.results = None
        self.test_run_scheduler = test_run_scheduler
        self.remote_debug_params = remote_debug_params

    def results_did_become_available(self, results):
        logger.debug('results avail:')
        logger.debug(results)
        self.results = results

    async def run(self, tests):
        self.timeline.mark_event(f'Splitting into CPU tasks. Total tests to run: {len(tests)}')
        #

        self.tasks = self.test_run_scheduler.schedule_into_tasks(tests)
        self.timeline.mark_event('Creating tcp thread')

        loop = asyncio.get_event_loop()
        server = await loop.create_server(
            lambda: self.create_server_protocol(),
            '127.0.0.1')
        # Allocate dynamic port. Extract port number from socket.
        port = server.sockets[0].getsockname()[1]
        self.timeline.mark_event('Subprocess: starting...')

        logger.debug('Initializing subprocess...')
        child_processes = []
        for task in self.tasks:
            child_processes.append(self.create_child_subprocess(port, task))

        logger.debug(f'Waiting for startup of {len(self.tasks)} subprocess task(s)')
        subprocesses_results = await asyncio.gather(*child_processes)
        logger.debug('Startup complete')

        child_waiters = []
        for _ in subprocesses_results:
            child_waiters.append(_.wait())

        logger.debug('Begin waiting for subprocess completion...')
        timeout_reached = False
        try:
            await asyncio.wait_for(
                asyncio.gather(*child_waiters),
                timeout=self.timeout_if_non_debug()
            )
        except (asyncio.TimeoutError, asyncio.CancelledError) as e:
            timeout_reached = True
            logger.warning(f'Reached execution timeout of {self.timeout_if_non_debug()} seconds. ')
            for _ in subprocesses_results:
                try:
                    _.kill()
                except:
                    logger.warning('Cannot kill child runner process with, ignoring.')

        logger.debug('All subprocesses are completed')
        logger.debug(f'Waiting for completion from TCP server')
        demo_results = []
        try:
            demo_results = await asyncio.wait_for(
                asyncio.gather(*self.completion_futures, return_exceptions=True),
                timeout=self.timeout_if_non_debug()
            )
        except asyncio.TimeoutError as ex:
            print(ex)
            for _ in self.client_connections:
                _.force_close()
            for _ in child_processes:
                _.close()
                print(_)
            pass

        logger.debug(f'TCP ran to the end')
        logger.debug(f'1 - merge_task_results')
        _results = self.merge_task_results(demo_results)
        logger.debug(f'1 - done')

        server.close()

        logger.debug(f' ---- TCP server and child processes ran to the end')
        if timeout_reached:
            raise asyncio.TimeoutError('Test execution timeout.')
        return _results

    def create_child_subprocess(self, port, task):
        # sys.executable is a full path to python.exe (or ./python) in current virtual environment
        return asyncio.create_subprocess_exec(sys.executable, *self.get_command_line_for_child(port, task.id), cwd=config.working_directory)

    def timeout_if_non_debug(self) -> Optional[float]:
        if self.remote_debug_params.enabled:
            return None
        return self.timeout

    def get_command_line_for_child(self, port, task_id):
        results = []
        results.append(f'{config.engine_directory}{os.sep}pycrunch{os.sep}multiprocess_child_main.py')
        results.append(f'--engine={config.runtime_engine}')
        results.append(f'--port={port}')
        results.append(f'--task-id={task_id}')
        results.append(f'--load-pytest-plugins={str(config.load_pytest_plugins).lower()}')
        if self.remote_debug_params.enabled:
            results.append(f'--enable-remote-debug')
            results.append(f'--remote-debugger-port={self.remote_debug_params.port}')
        if config.enable_web_ui:
            # No need to collect performance metrics if there is no web-ui
            results.append(f'--collect-perf')

        return results

    def create_server_protocol(self):
        loop = asyncio.get_event_loop()
        completion_future = loop.create_future()
        self.completion_futures.append(completion_future)

        # self.client_futures.append(completion_future)
        protocol = TestRunnerServerProtocol(self.tasks, completion_future, self.timeline)
        self.client_connections.append(protocol) 
        return protocol

    def merge_task_results(self, demo_results):
        self.timeline.mark_event('merging task results')
        new_dict = dict()
        # here are multiple dictionaries coming from each subprocess:
        for results_by_fqn in demo_results:
            new_dict.update(results_by_fqn)

        self.timeline.mark_event('merging task results -- Completed')
        self.timeline.mark_event(f'Total results processed {len(new_dict)}')

        return new_dict


from collections import defaultdict
from typing import Dict, Any, Union, Set

from pycrunch.api.shared import file_watcher
from pycrunch.session import config


class FileWithCoverage:
    def __init__(self, filename, lines_covered):
        self.lines_covered = lines_covered
        self.filename = filename


class FileStatistics:
    filename: str
    lines_with_entrypoints: Dict[str, Set[str]]

    def __init__(self, filename):
        self.filename = filename
        # line by line, each line contains one or multiple tests


        # 1: [module1:test1, ...]
        # 2: [module1:test_1, module2:test_2]
        self.lines_with_entrypoints = defaultdict(set)

    def mark_lines(self, lines_covered, fqn):

        for possibly_oudated_line in self.lines_with_entrypoints:
            self.lines_with_entrypoints[possibly_oudated_line].discard(fqn)

        for line in lines_covered:
            self.lines_with_entrypoints[line].add(fqn)

    def clear_file_from_test(self, fqn):
        for possibly_stale_line in self.lines_with_entrypoints:
            self.lines_with_entrypoints[possibly_stale_line].discard(fqn)



class CombinedCoverage:
    # FQN -> Files touched during run - i.e.: [file1.py, file2.py]
    file_dependencies_by_tests: Dict[str, Set[str]]
    # filename.py -> set(fqn, fqn)
    dependencies: Dict[str, Set[str]]

    # filename -> FileStatistics
    files: Dict[str, FileStatistics]
    """
        files[] -> line 1 -> [test1, test2]
                   line 2 -> [test2]
    """
    def __init__(self):
        self.files = dict()

        # all files involved in execution of test.
        # FQN will end up showing in multiple files if dependent file was used during run
        self.dependencies = defaultdict(set)
        self.file_dependencies_by_tests = defaultdict(set)
        #  in format
        #   {
        #       module:test_name : { status:failed },
        #       module_1:test_name_1 : { status:success },
        #    }
        # self.aggregated_results = defaultdict(dict)
        pass

    def mark_dependency(self, filename: str, fqn: str):
        if fqn not in self.dependencies[filename]:
            self.dependencies[filename].add(fqn)

    def test_did_removed(self, fqn: str):
        # clear dependencies
        for filename in self.dependencies:
            self.dependencies[filename].discard(fqn)

        if fqn in self.file_dependencies_by_tests:
            del self.file_dependencies_by_tests[fqn]

        # remove line hits from all files
        for stale_file in self.files:
            self.ensure_file_statistics_exist(stale_file)
            statistics = self.files[stale_file]
            statistics.clear_file_from_test(fqn=fqn)



    def mark_coverage(self, fqn, filename, lines_covered, test_run):
        self.ensure_file_statistics_exist(filename)
        statistics = self.files[filename]
        statistics.mark_lines(lines_covered=lines_covered, fqn=fqn, )

    def ensure_file_statistics_exist(self, filename):
        if not filename in self.files:
            statistics = FileStatistics(filename=filename)
            self.files[filename] = statistics

    def add_multiple_results(self, results):
        # todo invalidate\remove outdated runs

        for fqn, test_run in results.items():
            test_run = test_run

            self.clean_coverage_in_stale_files(fqn, test_run)
            for file in test_run.files:

                file_with_coverage = FileWithCoverage(filename=file.filename, lines_covered=file.lines)
                self.mark_dependency(file_with_coverage.filename, test_run.test_metadata['fqn'])

                self.mark_coverage(fqn=fqn, filename=file_with_coverage.filename, lines_covered=file_with_coverage.lines_covered, test_run=test_run)

        file_watcher.watch(self.files.keys())



    def clean_coverage_in_stale_files(self, fqn, test_run):
        # if file was not hit at all, we need to clear combined coverage there from previous runs
        previous_files = set(self.file_dependencies_by_tests[fqn])
        files = set()
        for single_file_run in test_run.files:
            files.add(single_file_run.filename)

        diff = previous_files - files
        for stale_file in diff:
            self.ensure_file_statistics_exist(stale_file)
            statistics = self.files[stale_file]
            statistics.clear_file_from_test(fqn=fqn)

        self.file_dependencies_by_tests[fqn] = files


def serialize_combined_coverage(combined: CombinedCoverage):
    return [
        dict(
            filename=config.path_mapping.map_to_local_fs(x.filename),
            lines_with_entrypoints=compute_lines(x)) for x in combined.files.values()
    ]


def compute_lines(x):
    return {
        line_number: list(entry_points)
        for (line_number, entry_points) in x.lines_with_entrypoints.items()
    }


combined_coverage = CombinedCoverage()


def unload_candidates(new_modules : set):
    reload_this = set()
    for x in new_modules:
        if is_candidate_for_reload(x):
            reload_this.add(x)

    return reload_this

def is_candidate_for_reload(x):
    if x.startswith('_pytest') or x.startswith('py.'):
        return False
    return True


import asyncio
import logging
import typing

from pycrunch.api.shared import pipe
from pycrunch.pipeline import execution_pipeline
from pycrunch.pipeline.download_file_task import DownloadFileTask
from pycrunch.pipeline.run_test_task import RemoteDebugParams, RunTestTask
from pycrunch.runner.pipeline_dispatcher import dispather_thread
from pycrunch.session.state import engine
from pycrunch.shared.models import all_tests
from .. import version
from ..execution_watchdog.connection_watchdog import connection_watchdog
from ..execution_watchdog.tasks import TerminateTestExecutionTask
from ..execution_watchdog.execution_watchdog import watchdog_dispather_thread
from ..execution_watchdog.watchdog_pipeline import watchdog_pipeline
from ..session import config

if typing.TYPE_CHECKING:
    import socketio

from threading import Lock

thread_lock = Lock()
thread = None
watchdog_thread = None

logger = logging.getLogger(__name__)


def attach_message_handlers_to_sio(sio: "socketio.Server"):
    @sio.on('message')
    def handle_message(message):
        logger.debug('received message 2: ' + message)


    @sio.on('json')
    async def handle_json(json, smth):
        logger.debug('handle_json')
        # logger.debug(session['userid'])
        # url_for1 = url_for('my event', _external=True)
        # logger.debug('url + ' + url_for1)
        await pipe.push(event_type='connected', **{'data': 'Connected'})
        logger.debug('received json 2: ' + str(json))


    @sio.on('my event')
    async def handle_my_custom_event(sid, json):
        logger.debug('received json (my event 2): ' + str(json))
        if 'action' not in json:
            logger.debug('no action specified')

        action = json.get('action')
        if action == 'discovery':
            await engine.will_start_test_discovery()
        if action == 'run-tests' or action == 'debug-tests':
            if 'tests' not in json:
                logger.error('run-tests command received, but no tests specified')
                return
            logger.info('Running tests...')
            tests = json.get('tests')
            fqns = set()
            for test in tests:
                fqns.add(test['fqn'])

            tests_to_run = all_tests.collect_by_fqn(fqns)
            if action == 'debug-tests':
                debugger_port = json.get('debugger_port')
                debug_params = RemoteDebugParams(True, debugger_port)
            else:
                debug_params = RemoteDebugParams.disabled()

            execution_pipeline.add_task(RunTestTask(tests_to_run, debug_params))
        if action == 'load-file':
            if config.enable_web_ui:
                filename = json.get('filename')
                logger.debug('download_file ' + filename)
                #         return asynchronously
                execution_pipeline.add_task(DownloadFileTask(filename))
        if action == 'diagnostics':
            await engine.will_start_diagnostics_collection()
        if action == 'timings':
            await engine.will_send_timings()
        if action == 'pin-tests':
            # fqns = array of strings[]
            fqns = json.get('fqns')
            await engine.tests_will_pin(fqns)
        if action == 'unpin-tests':
            fqns = json.get('fqns')
            await engine.tests_will_unpin(fqns)
        if action == 'engine-mode':
            new_mode = json.get('mode')
            engine.engine_mode_will_change(new_mode)
        if action == 'watchdog-terminate':
            print('action == watchdog-terminate -> TerminateTestExecutionTask')
            watchdog_pipeline.add_task(TerminateTestExecutionTask())

    @sio.event
    async def connect(sid, environ):
        global thread
        global watchdog_thread
        logger.debug('Client test_connected')
        connection_watchdog.connection_established()
        await pipe.push(
            event_type='connected',
            **dict(
                data='Connected test_connected',
                engine_mode=engine.get_engine_mode(),
                version=version.version_info
            )
        )

        with thread_lock:
            if thread is None:
                thread = 1
                loop = asyncio.get_event_loop()
                loop.create_task(dispather_thread())

            if watchdog_thread is None:
                watchdog_thread = 1
                loop = asyncio.get_event_loop()
                loop.create_task(watchdog_dispather_thread())

    @sio.event
    def disconnect(sid):
        logger.debug('Client disconnected')
        connection_watchdog.connection_lost()

    # method body is to register functions upon @sio.event
    pass


import pprint

timeline = None

def trace(*args, **kwargs):
    global timeline
    # todo: do not override, existing, create a list of subscribers
    if not timeline:
        # print('trace called, but no timeline injected: ', kwargs)
        return

    timeline.record(*args, **kwargs)


def inject_timeline(new_timeline):
    global timeline
    # print(timeline)

    timeline = new_timeline
    # print(timeline)


allowed_types = [int, str, float, dict, type(None), bool]

class RecordedVariable:
    def __init__(self, name, value, timestamp):
        self.name = name
        self.value = self.safe_for_serialization_value(value)
        self.timestamp = timestamp

    def safe_for_serialization_value(self, value):
        if type(value) not in allowed_types:
            return pprint.pformat(value)

        return value

    def as_json(self):
        self.recurse_fix_dict(self.value)
        return dict(
            ts=self.timestamp,
            name=self.name,
            value=self.value,
        )

    def recurse_fix_dict(self, v: dict):
        if type(v) == dict:
            for k, inner in v.items():
                self.recurse_fix_dict(inner)
                v[k] = self.safe_for_serialization_value(inner)



class EmptyInsightTimeline:
    def as_json(self):
        return []

class InsightTimeline:
    # Timeline must be viewed as chrome performance timeline or jetbrains timeline profiler UI
    # Timeline represents state of application on each line (ideally)
    # currently - on each call of trace function
    # :type variables: List[RecordedVariable]

    def __init__(self, clock):
        self.variables = []
        self.start_timestamp = None
        self.clock = clock
        self.counter = 1

    def start(self):
        self.start_timestamp = self.clock.now()

    def as_json(self):
        results = []
        for v in self.variables:
            results.append(v.as_json())
        return results

    def record(self, *args, **kwargs):
        # print(kwargs)
        # print(vars())
        ts = self.clock.now()
        adjusted_time = self.adjust_to_timeline_start(ts)
        for key, value in kwargs.items():
            # print(key + ' - ' + str(value))
            self.variables.append(RecordedVariable(key, value, adjusted_time))
        for value in args:
            self.variables.append(RecordedVariable(str(self.counter), value, adjusted_time))
            self.counter += 1

    def adjust_to_timeline_start(self, ts):
        return ts - self.start_timestamp

    def make_safe_for_pickle(self):
        # this is last resort call,
        # consider avoiding it in first place to not lose performance
        import pickle
        for variable in self.variables:
            try:
                pickle.dumps(variable)
            except Exception as e:
                variable.value = \
                    f'This cannot be traced: {str(variable.value)}\n\nConsider removing this trace call for faster test execution.'



import io
import logging
import multiprocessing
from pathlib import Path
from typing import List, Optional

import yaml
from pycrunch.constants import CONFIG_FILE_NAME
from pycrunch.session.auto_configuration import AutoConfiguration

logger = logging.getLogger(__name__)


class NoPathMapping:
    def map_to_local_fs(self, filename):
        return filename

    def map_local_to_remote(self, filename):
        return filename


class PathMapping:
    def __init__(self, path_in_container, path_on_local_ide):
        self.path_on_local_ide = path_on_local_ide
        self.path_in_container = path_in_container

    def map_to_local_fs(self, filename: str):
        return filename.replace(self.path_in_container, self.path_on_local_ide, 1)

    def map_local_to_remote(self, filename):
        return filename.replace(self.path_on_local_ide, self.path_in_container, 1)

class PycrunchException(Exception):
    pass

class Configuration:
    def __init__(self):
        self.allowed_modes = ['auto', 'manual', 'pinned']
        self.discovery_exclusions = ()
        self.coverage_exclusions = []
        self.working_directory = Path('.')
        self.django_ready = False
        self.engine_directory = 'unknown'
        self.engine_mode = 'auto'
        self.deep_inheritance = False
        self.pinned_tests = set()
        self.cpu_cores = self.get_default_cpu_cores()
        self.multiprocessing_threshold = 5
        self.execution_timeout_in_seconds = 60
        self.runtime_engine = 'pytest'
        self.available_engines = ['simple', 'pytest', 'django']
        self.environment_vars = dict()
        self.load_pytest_plugins = False
        self.path_mapping = NoPathMapping()
        self.enable_asyncio_debug = False
        self.enable_web_ui = False

    def runtime_engine_will_change(self, new_engine):
        self.throw_if_not_supported_engine(new_engine)
        self.runtime_engine = new_engine

    def prepare_django(self):
        # if not self.django_ready:
        if self.runtime_engine == 'django' and not self.django_ready:
            logger.info('!!! Importing Django and calling django.setup')
            try:
                import django
                django.setup()
            except Exception as e:
                logger.exception('Failed to import or set up django! Are you sure it is django project?', exc_info=e)
            self.django_ready = True


    def throw_if_not_supported_engine(self, new_engine):
        if new_engine not in self.available_engines:
            raise Exception(f'engine {new_engine} not available. Possible options: {self.available_engines}')

    def set_engine_directory(self, engine_directory):
        self.engine_directory = engine_directory

    def load_runtime_configuration(self):
        # This will output exact location of config file
        print(str(self.configuration_file_path().absolute()))
        print(self.configuration_file_path())

        auto_config = AutoConfiguration(self.configuration_file_path())
        auto_config.ensure_configuration_exist()

        try:
            with io.open(self.configuration_file_path(), encoding='utf-8') as f:
                self._load_config_now(f)
        except Exception as e:
            self.raise_configuration_error(e)

    def watch_for_config_changes(self):
        from pycrunch.api.shared import file_watcher
        logger.debug('watch_for_config_changes')
        file_watcher.watch([str(self.configuration_file_path().absolute())])

    def raise_configuration_error(self, e):
        print("Exception during processing configuration\n" + str(e))
        raise PycrunchException('configuration parse failed', e)

    def _load_config_now(self, f):
        x = yaml.safe_load(f)
        discovery = x.get('discovery', None)
        if discovery:
            exc = discovery.get('exclusions', None)
            if not hasattr(exc, "__len__"):
                raise Exception('.pycrunch-config.yaml: discovery->exclusions should be array')
            self.discovery_exclusions = tuple(exc)
        engine_config = x.get('engine', None)
        if engine_config:
            self._load_runtime_configuration_engine(engine_config)
        pinned_tests = x.get('pinned-tests', None)
        if pinned_tests:
            self.apply_pinned_tests(pinned_tests)
        additional_env = x.get('env', None)
        if additional_env:
            self.apply_additional_env(additional_env)
        path_mapping = x.get('path-mapping', None)
        if path_mapping:
            self.apply_path_mapping(path_mapping)

        self.apply_coverage_exclusions(x.get('coverage-exclusions', None))


    def _load_runtime_configuration_engine(self, engine_config):
        runtime_engine_name = engine_config.get('runtime', None)
        if runtime_engine_name:
            self.runtime_engine_will_change(runtime_engine_name)
        cpu_cores = engine_config.get('cpu-cores', None)
        if cpu_cores:
            self.cpu_cores_will_change(cpu_cores)

        enable_web_ui = engine_config.get('enable-web-ui', None)
        if enable_web_ui:
            if isinstance(enable_web_ui, bool):
                self.enable_web_ui = enable_web_ui
            else:
                print('engine: enable-web-ui parameter should be a boolean')
        multiprocess_threshold = engine_config.get('multiprocessing-threshold', None)
        if multiprocess_threshold:
            self.multiprocess_threshold_will_change(multiprocess_threshold)
        self.load_pytest_plugin_config(engine_config)


        # this is in seconds
        execution_timeout = engine_config.get('timeout', None)
        if execution_timeout is not None:
            self.execution_timeout_will_change(execution_timeout)
        self.deep_inheritance_will_change(engine_config)

    def configuration_file_path(self):
        return self.working_directory.joinpath(CONFIG_FILE_NAME)

    def runtime_mode_will_change(self, runtime_mode):
        self.throw_if_mode_not_supported(runtime_mode)
        print(f'Engine execution mode will change from {self.engine_mode} to {runtime_mode}')
        self.engine_mode = runtime_mode

    def throw_if_mode_not_supported(self, runtime_mode):
        if runtime_mode not in self.allowed_modes:
            raise Exception(f"runtime mode {runtime_mode} not supported. Available options are: {self.allowed_modes}")

    def load_pytest_plugin_config(self, engine_config):
        node: Optional[str] = engine_config.get('load-pytest-plugins', None)
        if node is not None:
            if type(node) == bool:
                self.load_pytest_plugins = node

    def deep_inheritance_will_change(self, engine_config):
        deep_inheritance: str = engine_config.get('deep-inheritance', None)
        if deep_inheritance is not None:
            if type(deep_inheritance) == bool:
                self.deep_inheritance = deep_inheritance

    def execution_timeout_will_change(self, new_timeout: float):
        if new_timeout < 0:
            logger.error(f'Execution timeout of {new_timeout} not valid. Fallback to default 60 sec. Please use positive numbers')
            return

        print(f'Using custom execution timeout of {new_timeout} seconds (default - 60 seconds)')
        self.execution_timeout_in_seconds = new_timeout

    def save_pinned_tests_config(self, fqns):
        self.apply_pinned_tests(fqns)
        existing_config = None
        with io.open(self.configuration_file_path(), encoding='utf-8', mode='r') as f:
            existing_config = yaml.safe_load(f)

        existing_config['pinned-tests'] = list(fqns)
        with io.open(self.configuration_file_path(), encoding='utf-8', mode='w') as f:
            yaml.dump(existing_config, f, default_flow_style=False)

    def is_test_pinned(self, fqn):
        return fqn in self.pinned_tests

    def apply_pinned_tests(self, pinned_tests):
        self.pinned_tests = set()
        for fqn in pinned_tests:
            self.pinned_tests.add(fqn)

    def apply_additional_env(self, additional_env):
        print('Using additional env from config:')
        print(additional_env)
        for env in additional_env:
            self.environment_vars[env] = additional_env[env]

    def apply_path_mapping(self, path_mapping):
        print('custom path map')
        for p in path_mapping:
            # only one for now
            self.path_mapping = PathMapping(p, path_mapping[p])

    def cpu_cores_will_change(self, cpu_cores):
        self.cpu_cores = cpu_cores
        pass

    def get_default_cpu_cores(self):
        cores = multiprocessing.cpu_count()
        if cores > 8:
            return 4

        if cores <= 2:
            return 1

        return round(cores / 2)

    def get_execution_timeout(self) -> Optional[float]:
        if self.execution_timeout_in_seconds == 0:
            return None
        return self.execution_timeout_in_seconds

    def multiprocess_threshold_will_change(self, multiprocessing_threshold):
        self.multiprocessing_threshold = multiprocessing_threshold

    def apply_coverage_exclusions(self, exclusions: Optional[List[str]]):
        if exclusions is None:
            return

        if not hasattr(exclusions, "__len__"):
            raise Exception('.pycrunch-config.yaml: apply_coverage->exclusions should be array')

        self.coverage_exclusions = list((f'*{_}*' if not _.endswith('.py') else f'*{_}' for _ in exclusions))


class _Lazy(object):
  _config_instance: Optional[Configuration]

  def __init__(self):
    self.download = None
    self._config_instance = None


  @property
  def config(self):
    if not self._config_instance:
      self._config_instance = Configuration()
    return self._config_instance

  def __getattr__(self, name):
    return self.config.__getattribute__ (name)


config = _Lazy() #  type: Optional[Configuration]





from pathlib import Path

from pycrunch.discovery.ast_discovery import AstTestDiscovery
from pycrunch.session.configuration import Configuration
import pycrunch.tests.dogfood.test_double_inheritance as double_inh

def test_simple_discovery():
    actual = run_dogfood_discovery()

    found_flag = False
    for t in actual.tests:
        if t.filename.endswith('test_discovery_specs_demo.py'):
            if t.name == 'test_regular':
                found_flag = True
                break

    assert found_flag


class TestProblem():
    def test_sample(self):
        assert 1 == 1


def test_double_inheritance():
    search_only_in = [double_inh.__file__]
    configuration = Configuration()
    configuration.deep_inheritance = True
    actual = run_dogfood_discovery(search_only_in, configuration)
    test_names = list(map(lambda _: _.name, actual.tests))

    assert 'DoublyInheritedScenario::test_1' in test_names


def test_only_methods_are_discovered_not_variables():
    actual = run_dogfood_discovery()
    test_names = list(map(lambda _: _.name, actual.tests))
    assert len(test_names) > 0
    assert 'test_variable' not in test_names

def test_classes_with_unit_tests_are_discoverable():
    actual = run_dogfood_discovery()
    test_names = list(map(lambda _: _.name, actual.tests))
    assert len(test_names) > 0
    assert 'SomeClassInhereted::test_1' in test_names
    assert 'TestX::test_1' in test_names
    assert 'MyClass::test_method1' in test_names
    assert 'MyClass::test_method2' in test_names
    assert 'TestForDummies::test_method1' in test_names
    assert 'TestForDummies::test_method2' in test_names
    assert 'TestForDummies::helper_method' not in test_names


def run_dogfood_discovery(search_only_in=None, config=None):
    root_folder = Path('.')
    current_folder = root_folder.joinpath('pycrunch', 'tests', 'dogfood').absolute()
    sut = AstTestDiscovery(str(root_folder.absolute()), config if config else Configuration())

    actual = sut.find_tests_in_folder(
        str(current_folder.absolute()),
        search_only_in=search_only_in
    )
    return actual





import ast
import logging
import sys
from pathlib import Path
from typing import List

from pycrunch.discovery.simple import TestSet, TestsInModule
from pycrunch.introspection.clock import Clock
from pycrunch.session import config
from pycrunch.session.file_map import test_map

logger = logging.getLogger(__name__)


class AstTestDiscovery:
    def __init__(self, root_directory=None, configuration=None):
        self.root_directory = root_directory
        self.time_spent = 0.0
        self.clock = Clock()

        self.configuration = configuration if configuration is not None else config

    def find_tests_in_folder(self, folder, search_only_in=None):
        import pathlib
        import sys

        if not self.root_directory:
            MODULE_DIR = folder
        else:
            MODULE_DIR = self.root_directory

        logger.debug(sys.path)
        logger.debug(f'MODULE_DIR {MODULE_DIR}')
        logger.debug(f'Discovering tests in folder {folder}')

        if not MODULE_DIR in sys.path:
            sys.path.insert(0, MODULE_DIR)
            logger.debug(f'after append')
            logger.debug(sys.path)
        # The directory containing your modules needs to be on the search path.

        # Get the stem names (file name, without directory and '.py') of any
        # python files in your directory, load each module by name and run
        # the required function.

        parent_path = pathlib.Path(MODULE_DIR)
        folder_path = pathlib.Path(folder)

        py_files = folder_path.glob('**/*.py')
        test_set = TestSet()

        for py_file in py_files:
            if search_only_in:
                if str(py_file) not in search_only_in:
                    continue

            current_file_path = py_file.relative_to(parent_path)
            if self.is_excluded_via_configuration(current_file_path):
                continue

            module_name = self.compute_module_name_from_path(current_file_path)
            if not self.is_module_with_tests(module_name):
                continue

            try:
                logger.debug('Compiling ' + module_name)
                ast_tree = self.load_syntax_tree_from(py_file)
                tests_found = self.load_tests_from_ast_representation(ast_tree)
            except Exception as ex:
                logger.error(f'Failed to parse ast of `{current_file_path}`')
                logger.exception(
                    f' parse {current_file_path} failed with exception: ' + str(ex)
                )
                continue

            filename = str(py_file)

            test_map.did_found_tests_in_file(filename, tests_found, module_name)
            test_set.add_module(TestsInModule(filename, tests_found, module_name))

            logger.debug(f'tests found: {tests_found}')

        logger.info(f' Time spent parsing AST : {round(self.time_spent, 3)} seconds')

        return test_set

    def load_syntax_tree_from(self, file_name) -> ast.Module:
        start = self.clock.now()
        contents = self.read_file(file_name)
        parse = ast.parse(contents, Path(file_name).name)
        end = self.clock.now()
        self.time_spent += end - start
        return parse

    def load_tests_from_ast_representation(self, ast_tree: ast.Module) -> List[str]:
        results = []

        def process_function_def(func: ast.FunctionDef) -> List[str]:
            if self.looks_like_test_name(func.name):
                return [func.name]

            return []

        def process_class_def(class_ast: ast.ClassDef) -> List[str]:
            class_results = []
            class_ast_name = class_ast.name

            c2 = self.looks_like_test_class(class_ast.name)
            if not c2:
                c2 = self.is_subclass_of_unittest(ast_tree, class_ast)

            if c2:
                for member in class_ast.body:
                    # TODO nested classes
                    if type(member) != ast.FunctionDef:
                        continue
                    member: ast.FunctionDef
                    if self.looks_like_test_name(member.name):
                        class_results.append(f"{class_ast_name}::{member.name}")
            return class_results

        mapping = {
            ast.FunctionDef: process_function_def,
            ast.ClassDef: process_class_def,
        }
        for node in ast_tree.body:
            walk_function = mapping.get(type(node))
            if not walk_function:
                continue
            values = walk_function(node)
            results.extend(values)

        return results

    @staticmethod
    def read_file(file_name: str) -> str:
        with open(file_name, "r") as f:
            return f.read()

    def compute_module_name_from_path(self, current_file_path):
        if len(current_file_path.parts) > 1:
            module_name = (
                str.join('.', current_file_path.parts[:-1])
                + '.'
                + current_file_path.stem
            )
        else:
            module_name = current_file_path.stem
        return module_name

    def is_excluded_via_configuration(self, current_file_path):
        s = str(current_file_path)
        x = False
        if s.startswith(self.configuration.discovery_exclusions) or s.endswith(
            self.configuration.discovery_exclusions
        ):
            x = True
        return x

    def is_module_with_tests(self, module_name):
        # Todo take pytest configs into account
        module_short_name = module_name.split('.')[-1]
        return module_short_name.startswith(
            ('test_', 'tests_')
        ) or module_short_name.endswith(('_test', 'tests', '_tests'))

    def looks_like_test_name(self, v):
        return v.startswith('test_') or v.endswith('_test')


    def looks_like_test_class(self, name: str) -> bool:
        return name.startswith('Test') or name.endswith('Test')

    def is_subclass_of_unittest(self, ast_module: ast.Module, class_ast: ast.ClassDef) -> bool:
        def search_via_deep_inheritance_analysis():
            if not self.configuration.deep_inheritance:
                return False

            # Compile and execute module
            code = compile(ast_module, filename='dummy', mode='exec')
            namespace = {}
            exec(code, namespace)

            # Check inheritance hierarchy
            may_null = namespace.get(class_ast.name)
            if may_null is None:
                return False
            may_be_not_loaded = sys.modules.get("unittest")
            is_unit_test_sub = may_be_not_loaded and issubclass(may_null, may_be_not_loaded.TestCase)

            return is_unit_test_sub

        available_subclasses = ['TestCase']

        for _ in class_ast.bases:
            type_of_base = type(_)
            if type_of_base == ast.Name:
                possible_none = getattr(_, 'id')
            elif type_of_base == ast.Attribute:
                possible_none = getattr(_, 'attr')
            else:
               continue

            if possible_none in available_subclasses:
                return True

        deep_inheritance_result = False
        try:
            deep_inheritance_result = search_via_deep_inheritance_analysis()
        except Exception as e:
            hint = 'You can disable this behaviour via `engine->deep-inheritance: false` configuration in .pycrunch-config.yaml'
            logger.warning(f'Failed to compile module to search via deep_inheritance_analysis for {class_ast.name}; Error: {e};\n{hint}')
        return deep_inheritance_result


# for basecls in inspect.getmro(self.obj.__class__):
#            dicts.append(basecls.__dict__)



import asyncio
import os
from typing import Dict, Any, Optional

from pycrunch.api import shared
from pycrunch.api.serializers import CoverageRun
from pycrunch.crossprocess.multiprocess_test_runner import MultiprocessTestRunner
from pycrunch.introspection.clock import clock
from pycrunch.introspection.history import execution_history
from pycrunch.introspection.timings import Timeline
from pycrunch.pipeline.abstract_task import AbstractTask
from pycrunch.runner.execution_result import ExecutionResult
from pycrunch.scheduling.scheduler import TestRunScheduler
from pycrunch.session.combined_coverage import combined_coverage, serialize_combined_coverage
from pycrunch.session.state import engine
from pycrunch.session import config

import logging

from pycrunch.execution_watchdog.tasks import TestExecutionBeginTask, TestExecutionEndTask
from pycrunch.execution_watchdog.execution_watchdog import termination_event
from pycrunch.execution_watchdog.watchdog_pipeline import watchdog_pipeline

logger = logging.getLogger(__name__)


class TestRunStatus:
    def __init__(self, status, results = None):
        """
        :type state: str
        :type results: Dict[str, CoverageRun]
        """
        if results is not None:
            logger.debug('results are not none')
        if results is None:
            logger.error('!!! None in results')
            results = dict()
        self.status = status
        self.results = results

    def is_failed(self):
        return self.status != 'success'


class RemoteDebugParams:
    def __init__(self, enabled: bool, port: Optional[int] = None):
        self.port = port
        self.enabled = enabled

    @classmethod
    def disabled(cls):
        return RemoteDebugParams(False)

class RunTestTask(AbstractTask):
    def __init__(self, tests, remote_debug_params: RemoteDebugParams):
        self.remote_debug_params = remote_debug_params
        self.timestamp = clock.now()
        self.tests = tests
        self.results = None
        self.timeline = Timeline('run tests')
        self.timeline.start()

    async def run(self):
        """
            Here we run multiple tests at once using one or multiple processes
        """
        self.timeline.mark_event('run')
        watchdog_pipeline.add_task(TestExecutionBeginTask(len(self.tests)))
        socket_notification_task = asyncio.ensure_future(engine.tests_will_run(self.tests))

        converted_tests = self.get_converted_test_list()
        runner = self.create_test_runner()

        # while not cancelled
        runner_task = asyncio.ensure_future(runner.run(tests=converted_tests))
        run_results_compound = await self.wait_with_cancellation(runner_task)
        if run_results_compound.is_failed():
            failure_reason = self.user_friendly_error_message(run_results_compound.status)

            for _ in converted_tests:
                candidate_fqn = _['fqn']
                cov_run = CoverageRun(candidate_fqn, -1, None, execution_result=ExecutionResult.create_failed_with_reason(failure_reason))
                run_results_compound.results[candidate_fqn] = cov_run


        run_results = run_results_compound.results

        self.timeline.mark_event('before tests_did_run')

        # asynchronously send message over websocket
        # Line bellow communicates test statuses as a side effect
        async_tasks_post = [
            engine.tests_did_run(run_results)
        ]

        self.post_process_combined_coverage(run_results)

        self.timeline.mark_event('Sending: test_run_completed event')
        # todo: i'm only using `filename` in connector, why bother with everything?
        cov_and_run_details_to_send = dict(all_runs=self.convert_result_to_json(run_results))

        async_tasks_post.append(shared.pipe.push(
            event_type='test_run_completed',
            coverage=cov_and_run_details_to_send,
            timings=dict(start=self.timestamp, end=clock.now()),
        ))

        self.timeline.mark_event('Started combined coverage serialization')
        serialized = serialize_combined_coverage(combined_coverage)
        self.timeline.mark_event('Completed combined coverage serialization')

        self.timeline.mark_event('Sending: combined coverage over WS')
        async_tasks_post.append(
            shared.pipe.push(
                event_type='combined_coverage_updated',
                combined_coverage=serialized,
                aggregated_results=engine.all_tests.legacy_aggregated_statuses(),
                timings=dict(start=self.timestamp, end=clock.now())
            )
        )

        self.timeline.mark_event('Waiting until post-processing tasks are completed')
        await asyncio.gather(*async_tasks_post)
        watchdog_pipeline.add_task(TestExecutionEndTask())

        self.timeline.mark_event('Send: done, stopping timeline')

        self.timeline.stop()
        execution_history.save(self.timeline)

    def user_friendly_error_message(self, status: str):
        failure_reason = 'epic fail'
        if status == 'terminated':
            failure_reason = 'Test execution terminated by user.'
        if status == 'timeout':
            line1 = f'Timeout of {config.execution_timeout_in_seconds} seconds reached while waiting for test execution to complete.'
            line2 = f'Consider increasing it in .pycrunch-config.yaml, e.g.:'
            line3 = f'{os.linesep}engine:{os.linesep}    timeout: 999{os.linesep}'
            line4 = f'Setting it to zero will wait forever.{os.linesep}'
            line5 = 'https://pycrunch.com/docs/configuration-file'
            failure_reason = os.linesep.join([line1, line2, line3, line4, line5])
        return failure_reason

    def convert_result_to_json(self, run_results):
        results_as_json = dict()
        for k, v in run_results.items():
            results_as_json[k] = v.as_json()
        return results_as_json

    def post_process_combined_coverage(self, run_results):
        if self.remote_debug_params.enabled:
            self.timeline.mark_event('Postprocessing: combined coverage will not be recomputed.')
            return

        self.timeline.mark_event('Postprocessing: combined coverage, line hits, dependency tree')
        combined_coverage.add_multiple_results(run_results)
        self.timeline.mark_event('Postprocessing: completed')

    def create_test_runner(self):
        self.timeline.mark_event('before running tests')

        runner = MultiprocessTestRunner(
            timeout=config.get_execution_timeout(),
            timeline=self.timeline,
            test_run_scheduler=TestRunScheduler(
                cpu_cores=config.cpu_cores,
                threshold=config.multiprocessing_threshold
            ),
            remote_debug_params=self.remote_debug_params,
        )
        return runner

    async def wait_with_cancellation(self, runner_task: asyncio.Future) -> TestRunStatus:
        try:
            # Here we wait for the first event, which may be:
            # 1. Watchdog termination signal
            # 2. Test runner events
            #  2.1 Run to end
            #  2.2 Timeout during run
            waited = await asyncio.wait([termination_event.wait(), runner_task], return_when=asyncio.FIRST_COMPLETED)
            if runner_task.done():
                return TestRunStatus('success', runner_task.result())
            if termination_event.is_set():
                logger.warning('Looks like task was cancelled by user...')
                runner_task.cancel()
                return TestRunStatus('terminated')
        except asyncio.TimeoutError:
            logger.warning('Timeout reached while waiting for tests...')
            return TestRunStatus('timeout')


    def get_converted_test_list(self):
        converted_tests = list()
        for test in self.tests:
            # todo why is this line exist?
            converted_tests.append(
                dict(fqn=test.discovered_test.fqn, filename=test.discovered_test.filename, name=test.discovered_test.name, module=test.discovered_test.module, state='converted'))
        return converted_tests

# https://stackoverflow.com/questions/45369128/python-multithreading-queue



from unittest import TestCase, skip


class TestWithSkips(TestCase):
    def setUp(self) -> None:
        # by some reason skipped test without reason still runs setup
        pass

    @skip
    def test_should_be_skipped(self):
        print('this should show as success in pycrunch report')

    @skip('reason')
    def test_should_be_skipped_with_reason(self):
        # py.test reporter seems to use different algo to skip test with reason and without
        # make sure both of them works

        print('this should show as success in pycrunch report')

    def test_alongside_with_skipped(self):
        assert True




from collections import deque


def serialize_intervals(root_interval):
    result = []
    intervals_dict = []
    events = []
    for interval in root_interval.intervals:
        intervals_dict.append(serialize_intervals(interval))
    for evt in root_interval.events:
        events.append(dict(name=evt.name, timestamp=evt.relative_timestamp()))
    result.append(dict(name=root_interval.name, events=events, intervals=intervals_dict, start=root_interval.started_at, end=root_interval.stopped_at, duration=root_interval.duration()))
    return result


class ExecutionHistory:
    def __init__(self):
        self.timelines = deque(maxlen=100)

    def save(self, timeline):
        self.timelines.append(timeline)

    def to_json(self):
        results = []
        my_copy = self.timelines.copy()
        for x in my_copy:
            results.append(dict(timeline_name=x.name, duration=x.duration(), intervals=serialize_intervals(x.root)))
        return dict(results=results)


execution_history = ExecutionHistory()


from pycrunch.insights.variables_inspection import EmptyInsightTimeline


class ErrorRecord: 
    def __init__(self, etype, value, current_traceback):
        self.current_traceback = current_traceback
        self.value = value
        self.etype = etype


failed_status = 'failed'
success_status = 'success'
queued_status = 'queued'

class ExecutionResult:
    def __init__(self):
        self.captured_output = None
        self.status = 'pending'
        self.intercepted_exception = None
        self.state_timeline = EmptyInsightTimeline()

    def record_exception(self, etype, value, current_traceback):
        self.intercepted_exception = ErrorRecord(etype=etype, value=value, current_traceback=current_traceback)

    def run_did_fail(self):
        self.status = failed_status

    def run_did_queued(self):
        self.status = queued_status

    def run_did_succeed(self):
        self.status = success_status

    def output_did_become_available(self, captured_output):
        self.captured_output = captured_output

    def state_timeline_did_become_available(self, state_timeline):
        self.state_timeline = state_timeline
    
    @classmethod
    def create_failed_with_reason(cls, reason):
        """

        :type reason: str
        """
        x = cls()
        x.run_did_fail()
        x.output_did_become_available(reason)
        return x



import setuptools
from distutils.core import setup

setup(name='pycrunch-engine',
      version='1.4',
      description='Automatic Test Runner Engine',
      url='http://github.com/gleb-sevruk/pycrunch-engine',
      author='Gleb Sevruk',
      author_email='support@pycrunch.com',
      license='libpng',
      keywords="tdd unit-testing test runner",
      packages=setuptools.find_packages(),
      project_urls={
              'Documentation': 'https://pycrunch.com/docs',
              'Funding': 'https://pycrunch.com/donate',
              'PyCharm Plugin': 'https://plugins.jetbrains.com/plugin/13264-pycrunch--live-testing',
              'Source': 'https://github.com/gleb-sevruk/pycrunch-engine',
              'Tracker': 'https://github.com/gleb-sevruk/pycrunch-engine/issues',
      },
      setup_requires=['wheel'],
      entry_points={
              'console_scripts': ['pycrunch-engine=pycrunch.main:run'],
      },
      install_requires=[
          'pytest',
          'coverage>4',
          'PyYAML',
          'watchdog',
          'python-socketio>=4,<5',
          'aiohttp',
      ],
      include_package_data=True,
      zip_safe=False)



import datetime
import os

from pycrunch.api.serializers import CoverageRun
from pycrunch.child_runtime.coverage_hal import CoverageAbstraction
from pycrunch.insights.variables_inspection import InsightTimeline, inject_timeline
from pycrunch.runner.execution_result import ExecutionResult

DISABLE_COVERAGE = False

class TestRunner:
    def __init__(self, runner_engine, timeline, coverage_exclusions, child_config):
        self.runner_engine = runner_engine
        self.timeline = timeline
        self.child_config = child_config
        self.coverage_exclusions = coverage_exclusions

    def run(self, tests):
        self.timeline.mark_event('Run: inside run method')
        from pycrunch.introspection.clock import clock
        from pycrunch.runner.interception import capture_stdout
        from pycrunch.shared.primitives import TestMetadata
        self.timeline.mark_event('Run: inside run method - imports complete')

        results = dict()
        for test_to_run in tests:
            self.timeline.begin_nested_interval(f'Running test {test_to_run.get("fqn", "unknown")}')

            # record traced variables
            state_timeline = InsightTimeline(clock=clock)
            state_timeline.start()
            inject_timeline(state_timeline)

            metadata = TestMetadata(**test_to_run)
            try:
                # TODO: Check if starting coverage AFTER pytest.main,
                #    before test_method enter, improves time in magnitudes.
                #  ---
                #    checked, there are 2x improvement for small files (0.06 vs 0.10, but still
                #      slow as before on 500+ tests in one file
                should_disable_coverage = DISABLE_COVERAGE
                if self.child_config.enable_remote_debug:
                    should_disable_coverage = True
                cov = CoverageAbstraction(should_disable_coverage, self.coverage_exclusions, self.timeline)
                cov.start()

                with capture_stdout() as get_value:
                    time_start = clock.now()
                    self.timeline.mark_event('About to start test execution')
                    execution_result = self.runner_engine.run_test(metadata)
                    self.timeline.mark_event('Test execution complete, postprocessing results')
                    time_end = clock.now()
                    time_elapsed = time_end - time_start

                    cov.stop()

                    _now = datetime.datetime.now()

                    print(f'{os.linesep}at {_now.strftime("%X.%f")[:-3]} {_now.strftime("%x")}')

                    captured_output = get_value()
                    self.timeline.mark_event('Received captured output')

                    execution_result.output_did_become_available(captured_output)
                    execution_result.state_timeline_did_become_available(state_timeline)

                    self.timeline.mark_event('Before coverage serialization')
                    coverage_for_run = self.serialize_test_run(cov, metadata.fqn, time_elapsed, test_metadata=test_to_run, execution_result=execution_result)
                    self.timeline.mark_event('After coverage serialization')
            except Exception as e:
                # Here is most likely exception in the engine itself.
                self.timeline.mark_event('Test execution exception.')
                import sys
                tb = self.get_detailed_traceback(metadata.fqn)
                print(tb, file=sys.__stdout__)
                from pycrunch.api.serializers import CoverageRun
                result = ExecutionResult.create_failed_with_reason(tb)
                # inject fake run to not crash entire pipeline
                coverage_for_run = CoverageRun(metadata.fqn, -1, test_to_run, execution_result=result)

                # logger.exception('error during run', exc_info=e)
                
            results[metadata.fqn] = coverage_for_run
            self.timeline.end_nested_interval()
        return results

    def serialize_test_run(self, cov, fqn, time_elapsed, test_metadata, execution_result):
        """
        :type cov: CoverageAbstraction
        :param time_elapsed: float
        :param fqn: str
        """
        run_results = CoverageRun(fqn, time_elapsed, test_metadata, execution_result)
        run_results.store_files_coverage(cov.parse_all_hit_lines())
        return run_results


    def get_detailed_traceback(self, fqn):
        """

        :type fqn: str
        :return str
        """
        import traceback
        from io import StringIO
        out = StringIO()
        print('----', file=out)
        print('Error in PyCrunch subprocess executor ', file=out)
        print(f'   while running test `{fqn}`', file=out)
        print('----', file=out)
        traceback.print_exc(file=out)
        return out.getvalue()



